"""Iniitialize the KIE task and load cached values"""
import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Tuple, Type
from urllib.parse import urljoin

import mlflow
from databricks.sdk import WorkspaceClient  # pylint: disable = ungrouped-imports
from mlflow.entities import Experiment
from pydantic import BaseModel
from pyspark.sql import DataFrame

from databricks.kie.data_utils import split_labeled_data, split_unlabeled_data
from databricks.kie.kie_schema import ModelFactory
from databricks.kie.kie_state import KIEState
from databricks.kie.prompt_builder import PromptBuilder
from databricks.kie.task_spec import KIETaskSpec
from databricks.model_training.api.utils import get_browser_url, get_schema_from_table, get_spark

_SHARE_KEY = "kie_view"


def get_table_if_exists(path: str) -> Optional[DataFrame]:
    # Load files from cache if they exist
    wc = WorkspaceClient()
    try:
        if wc.tables.exists(path).table_exists:
            spark = get_spark()
            return spark.read.table(path)
    except:  # pylint: disable=bare-except
        return None


def get_mlflow_experiment_link(experiment_id: str) -> str:
    """
    Get a link for the MLflow experiment using the given link text.
    This link will open the MLflow experiment page in a new tab.
    """
    host = get_browser_url()
    path = urljoin(host, f'/ml/experiments/{experiment_id}?viewStateShareKey={_SHARE_KEY}')
    return path


@dataclass
class CachePaths:
    """Holds all cache paths for a KIE task."""
    unlabeled_table: str
    labeled_table: str
    grounding_table: str
    val_table: str
    train_jsonl: str
    schema_json: str

    @classmethod
    def from_task_spec(cls, task_spec: KIETaskSpec, root_name: str) -> 'CachePaths':
        output_schema = get_schema_from_table(task_spec.output_table)
        cache_root = f"{output_schema}.{root_name}"

        return cls(unlabeled_table=f"{cache_root}_unlabeled",
                   labeled_table=f"{cache_root}_labeled",
                   grounding_table=f"{cache_root}_grounding",
                   val_table=f"{cache_root}_val",
                   train_jsonl=str(Path(task_spec.output_path) / f"{root_name}_train.jsonl"),
                   schema_json=str(Path(task_spec.output_path) / f"{root_name}_schema.json"))


class TaskInitializer:
    """Initializes a KIE task's state from a task spec"""

    def __init__(self, task_spec: KIETaskSpec):
        self.task_spec = task_spec
        self.spark = get_spark()

    def _setup_experiment(self) -> Experiment:
        """Setup and return MLflow experiment."""
        experiment = mlflow.get_experiment_by_name(self.task_spec.experiment_name)
        if experiment is None:
            experiment_id = mlflow.create_experiment(self.task_spec.experiment_name)
            experiment = mlflow.get_experiment(experiment_id)
            path = get_mlflow_experiment_link(experiment.experiment_id)
            print(f"🧪 Created a new MLflow experiment at: {path}")
        else:
            path = get_mlflow_experiment_link(experiment.experiment_id)
            print(f"🧪 Using MLflow experiment at: {path}")

        mlflow.set_experiment(experiment_id=experiment.experiment_id)
        return experiment

    def _get_response_format(self, cache_paths: CachePaths) -> Type[BaseModel]:
        """Get or create response format."""
        if os.path.exists(cache_paths.schema_json):
            return ModelFactory.from_file(cache_paths.schema_json)

        if self.task_spec.labeled_dataset and not self.task_spec.json_examples:
            labeled_df = self.spark.read.table(self.task_spec.labeled_dataset)
            rows = labeled_df.limit(10).collect()
            self.task_spec.json_examples = [
                json.loads(r[self.task_spec.labeled_dataset_output_json_column]) for r in rows
            ]

        return ModelFactory.from_examples(self.task_spec.json_examples, "response_format")

    def _setup_data_splits(self, cache_paths: CachePaths, num_grounding_samples: int,
                           num_val_samples: int) -> Tuple[Optional[DataFrame], DataFrame]:
        """Setup labeled and unlabeled data splits."""

        labeled_split_df = get_table_if_exists(cache_paths.labeled_table)
        if self.task_spec.labeled_dataset and labeled_split_df is None:
            labeled_df = self.spark.read.table(self.task_spec.labeled_dataset)
            labeled_split_df = split_labeled_data(labeled_df,
                                                  num_grounding_samples=num_grounding_samples,
                                                  num_val_samples=num_val_samples)
            labeled_split_df.write.mode("overwrite").saveAsTable(cache_paths.labeled_table)
            num_val_samples = num_grounding_samples = 0
            print(f"Sourcing labeled data from {self.task_spec.labeled_dataset}")

        unlabeled_split_df = get_table_if_exists(cache_paths.unlabeled_table)
        if unlabeled_split_df is None:
            unlabeled_split_df = split_unlabeled_data(self.task_spec.unlabeled_dataset,
                                                      num_val_samples=num_val_samples,
                                                      num_grounding_samples=num_grounding_samples)
            unlabeled_split_df.write.mode("overwrite").saveAsTable(cache_paths.unlabeled_table)

        print(f"Found {unlabeled_split_df.count()} unlabeled documents to extract")
        return labeled_split_df, unlabeled_split_df

    def initialize(self, num_grounding_samples: int, num_val_samples: int, num_fewshot_samples: int) -> KIEState:
        """Initialize KIE task state."""
        experiment = self._setup_experiment()
        root_name = os.path.basename(experiment.name).replace("-", "_")
        cache_paths = CachePaths.from_task_spec(self.task_spec, root_name)

        response_format = self._get_response_format(cache_paths)
        prompt_builder = PromptBuilder(response_format)

        labeled_split_df, unlabeled_split_df = self._setup_data_splits(cache_paths, num_grounding_samples,
                                                                       num_val_samples)

        grounding_df = get_table_if_exists(cache_paths.grounding_table)
        val_df = get_table_if_exists(cache_paths.val_table)

        return KIEState(
            experiment=experiment,
            prompt_builder=prompt_builder,
            ground_truth_prompt=prompt_builder.build_prompt(),
            zeroshot_prompt=prompt_builder.build_prompt(include_markdown=True),
            fewshot_prompt=None,
            grounding_table_path=cache_paths.grounding_table,
            val_table_path=cache_paths.val_table,
            train_jsonl_path=cache_paths.train_jsonl,
            schema_path=cache_paths.schema_json,
            requires_grounding=grounding_df is None,
            requires_val=val_df is None,
            requires_train_gen=not os.path.exists(cache_paths.train_jsonl),
            unlabeled_split_df=unlabeled_split_df,
            labeled_split_df=labeled_split_df,
            grounding_df=grounding_df,
            val_df=val_df,
            response_format=response_format,
            num_grounding_samples=num_grounding_samples,
            num_val_samples=num_val_samples,
            num_fewshot_samples=num_fewshot_samples,
            model_dfs=None,
        )



"""
A Databricks training run
"""

from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime, timezone
from http import HTTPStatus
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urljoin

import mlflow
from mcli.api.schema.generic_model import DeserializableModel, convert_datetime
from mcli.utils.utils_string_functions import camel_case_to_snake_case

from databricks.model_training.api.engine.engine import MAPIException
from databricks.model_training.api.utils import get_host_and_token_from_env, is_running_in_databricks_notebook
from databricks.model_training.types.run_status import RunStatus
from databricks.model_training.types.train_config import TrainConfig

PROGRESS = [
    '□□□□□□□□□□', '■□□□□□□□□□', '■■□□□□□□□□', '■■■□□□□□□□', '■■■■□□□□□□', '■■■■■□□□□□', '■■■■■■□□□□', '■■■■■■■□□□',
    '■■■■■■■■□□', '■■■■■■■■■□', '■■■■■■■■■■'
]

_MLFLOW_CLIENT = mlflow.MlflowClient(tracking_uri='databricks', registry_uri='databricks-uc')


def seconds_to_str(seconds: Union[int, float]) -> str:
    """Converts seconds to a human-readable string

    Args:
        seconds: Number of seconds

    Returns:
        Human-readable string
    """
    if seconds < 60:
        return f'{int(seconds)}s'
    elif seconds < 3600:
        return f'{int(seconds/60)}min'
    elif seconds < 24 * 3600:
        return f'{int(seconds/3600)}hr'
    else:
        return f'{int(seconds/3600/24)}d'


@dataclass()
class TrainingEvent(DeserializableModel):
    """ Finetuning Run Event """
    type: str
    time: datetime
    message: str

    @classmethod
    def from_mapi_response(cls, response: Dict[str, Any]) -> 'TrainingEvent':
        """Load the formatted run event from MAPI response.
        """
        args = {camel_case_to_snake_case(key): value for key, value in response.items()}
        return cls(
            type=args['event_type'],
            time=args['event_time'],
            message=args['event_message'],
        )

    def is_terminal_state(self):
        return self.type in ['COMPLETED', 'FAILED', 'STOPPED', 'CANCELED']


@dataclass()
class TrainingRun(DeserializableModel):
    """A Databricks training run

    Args:
        name: The name of the training run
        created_by: The user email of who created the run
        model: The model to finetune
        train_data_path: The path to the training data
        register_to: The location to the registered model
        submitted_experiment_path: The submitted MLflow experiment path
        task_type: The type of training task to run
        eval_data_path: The path to the eval data
        eval_prompts: The list of prompts to use for eval
        custom_weights_path: The path to a custom model checkpoint to use for training
        training_duration: The total duration of the training run
        learning_rate: The peak learning rate to use for training
        context_length: The maximum sequence length to use
        data_prep_cluster_id: The id of the cluster used for data prep
        created_at: The time the run was created
        started_at: The time the run was started
        estimated_end_time: The estimated time the run will complete
        completed_at: The time the run was completed
        details: The current run event
    """

    name: str
    status: RunStatus
    created_by: str

    # User inputs
    model: Optional[str] = None
    save_folder: Optional[str] = None
    train_data_path: Optional[str] = None
    register_to: Optional[str] = None
    submitted_experiment_path: Optional[str] = None
    task_type: Optional[str] = 'INSTRUCTION_FINETUNE'
    eval_data_path: Optional[str] = None
    eval_prompts: Optional[List[str]] = None
    custom_weights_path: Optional[str] = None
    training_duration: Optional[str] = None
    learning_rate: Optional[float] = None
    context_length: Optional[int] = None
    # we do not include validate_inputs as this is not sent to MAPI nor stored anywhere
    data_prep_cluster_id: Optional[str] = None

    # holds the current run event
    details: Optional[str] = None

    # Lifecycle
    created_at: Optional[datetime] = None
    started_at: Optional[datetime] = None
    estimated_end_time: Optional[datetime] = None
    formatted_eta: Optional[str] = None
    completed_at: Optional[datetime] = None

    # pylint: disable-next=import-outside-toplevel,cyclic-import
    from databricks.model_training.types.common import ObjectList

    # Details
    submitted_config: Optional[TrainConfig] = None
    events: Optional[ObjectList[TrainingEvent]] = None

    # MLflow
    experiment_id: Optional[str] = None
    run_id: Optional[str] = None

    _required_properties: Tuple[str, ...] = tuple([
        'id',
        'name',
        'status',
        'createdByEmail',
        'createdAt',
        'updatedAt',
    ])

    @property
    def run_progress(self) -> str:
        if self.status == RunStatus.RUNNING and self.started_at is not None and self.estimated_end_time is not None:
            total_time = (self.estimated_end_time - self.started_at).total_seconds()
            elapsed_time = (datetime.now(timezone.utc) - self.started_at).total_seconds()
            percentage = min(int((elapsed_time / total_time) * 100), 100)
            return f'{RunStatus.RUNNING} {PROGRESS[int(percentage / 10)]}({percentage}%)'
        return f'{self.status}'

    @property
    def formatted_events(self) -> str:
        return '' if self.events is None else '<br>'.join(
            ['<nobr>' + event.type + ' ' + str(event.time) + '</nobr>' for event in self.events])

    # hyperparameters: Hyperparameters ## todo

    @classmethod
    def from_mapi_response(cls, response: Dict[str, Any]) -> 'TrainingRun':
        """Load the training run from MAPI response.
        """
        missing = set(cls._required_properties) - set(response)
        if missing:
            raise MAPIException(
                status=HTTPStatus.BAD_REQUEST,
                message='Missing required key(s) in response to deserialize TrainingRun '
                f'object: {", ".join(missing)} ',
            )
        started_at = convert_datetime(response['startedAt']) if response.get('startedAt', None) else None
        completed_at = convert_datetime(response['completedAt']) if response.get('completedAt', None) else None
        estimated_end_time = convert_datetime(response['estimatedEndTime']) if response.get('estimatedEndTime',
                                                                                            None) else None

        args = {
            'name': response['name'],
            'created_at': convert_datetime(response['createdAt']),
            'started_at': started_at,
            'completed_at': completed_at,
            'status': RunStatus.from_string(response['status']),
            'details': response.get('reason', ''),
            'created_by': response['createdByEmail'],
            'estimated_end_time': estimated_end_time
        }

        if estimated_end_time is not None:
            args['estimated_end_time'] = estimated_end_time

        details = response.get('details', {})
        if len(details) > 0:
            args['model'] = details.get('model')
            args['save_folder'] = details.get('saveFolder')
            args['train_data_path'] = details.get('trainDataPath')
            args['eval_data_path'] = details.get('evalDataPath')
            args['eval_prompts'] = details.get('evalPrompts')
            args['custom_weights_path'] = details.get('customWeightsPath')
            args['training_duration'] = details.get('trainingDuration')
            args['learning_rate'] = details.get('learningRate')
            args['context_length'] = details.get('contextLength')
            experiment_tracker = details.get('experimentTracker')
            if experiment_tracker is None or experiment_tracker.get('mlflow') is None:
                raise MAPIException(
                    status=HTTPStatus.NOT_FOUND,
                    message=
                    'Missing MLflow experimentTracker, which is a required field to deserialize TrainingRun object',
                )
            mlflow_config = experiment_tracker.get('mlflow')
            args['register_to'] = mlflow_config.get('modelRegistryPath')
            args['submitted_experiment_path'] = mlflow_config.get('experimentPath')
            args['experiment_id'] = mlflow_config.get('mlflowExperimentId')
            args['run_id'] = mlflow_config.get('mlflowRunId')
            args['task_type'] = details.get('taskType')

            config_copy = deepcopy(details)
            # Remove events from details to keep only config properties
            if 'formattedFinetuningEvents' in config_copy:
                del config_copy['formattedFinetuningEvents']

            if 'dataPrepConfig' in config_copy:
                if 'clusterId' in config_copy['dataPrepConfig']:
                    args['data_prep_cluster_id'] = config_copy['dataPrepConfig']['clusterId']
                del config_copy['dataPrepConfig']
            args['submitted_config'] = TrainConfig.from_mapi_response(config_copy)

            formatted_training_events = sorted(
                [TrainingEvent.from_mapi_response(event) for event in details.get('formattedFinetuningEvents', [])],
                key=lambda x: x.time)

            # pylint: disable-next=import-outside-toplevel,cyclic-import
            from databricks.model_training.types.common import ObjectList, ObjectType

            # convert to ObjectList for better rendering when field referenced
            args['events'] = ObjectList(formatted_training_events, ObjectType.from_model_type(TrainingEvent))
            # Don't show ETA during checkpoint upload
            if args['estimated_end_time'] is not None:
                found = next((event for event in args['events'] if event.type == 'TRAIN_FINISHED'), None)
                if found is not None:
                    args['estimated_end_time'] = None

        if args['estimated_end_time'] is not None:
            time_left = (args['estimated_end_time'] - datetime.now(timezone.utc)).total_seconds()
            if time_left <= 0:
                args['estimated_end_time'] = None
            else:
                args['formatted_eta'] = seconds_to_str(time_left)

        return cls(**args)

    def refresh(self) -> 'TrainingRun':
        """Refetches the training run from the API

        Returns:
            TrainingRun: The updated training run
        """

        # pylint: disable-next=import-outside-toplevel,cyclic-import
        from databricks.model_training.api.foundation_model.get import get
        return get(self)

    def cancel(self) -> int:
        """Cancel the training run

        Returns:
            int: Will return 1 if the run was cancelled, 0 if it was already cancelled
        """

        # pylint: disable-next=import-outside-toplevel,cyclic-import
        from databricks.model_training.api.foundation_model.cancel import cancel
        return cancel(self)

    def get_events(self) -> ObjectList[TrainingEvent]:
        """Get events for the training run

        Returns:
            List[TrainingEvent]: A list of training run events. Each event has an event
                type, time, and message.
        """

        # pylint: disable-next=import-outside-toplevel,cyclic-import
        from databricks.model_training.api.foundation_model.get_events import get_events
        return get_events(self)

    def _repr_html_(self) -> str:
        """Display the event as HTML.
        """

        keys = [
            'name', 'status', 'details', 'formatted_events', 'model', 'task_type', 'learning_rate', 'training_duration',
            'train_data_path', 'eval_data_path', 'register_to', 'experiment_id', 'eval_prompts', 'custom_weights_path',
            'data_prep_cluster_id'
        ]

        attr_to_label = {key: ' '.join([word.title() for word in key.split('_')]) for key in keys}
        # If the experiment_id or run_id isn't set, we can't display the full path to the MLflow run.
        # These experiment and path fields are mutable, so we can't rely on them to be the same as when
        # the run was created.
        if self.experiment_id is None or self.run_id is None:
            attr_to_label['submitted_experiment_path'] = 'MLflow Experiment'
        else:
            attr_to_label['_get_mlflow_path_html_with_link'] = 'MLflow Run'
        # override title for events
        attr_to_label['formatted_events'] = 'Events'

        # pylint: disable-next=import-outside-toplevel,cyclic-import
        from databricks.model_training.types.common import generate_vertical_html_table
        return generate_vertical_html_table([self], attr_to_label)

    @property
    def experiment_path(self) -> str:
        """
        Get the experiment path for this run, not including run name (i.e. `/Users/username/experiment_name`).
        This uses the `experiment_id` attribute to get the latest experiment name, since experiment names
        are mutable.

        For runs created before 04/29/24, experiment id is not populated, so this returns the submitted experiment path.
        """
        if not self.experiment_id:
            if self.submitted_experiment_path is None:
                raise ValueError('Experiment ID and path are not set')
            return self.submitted_experiment_path
        return _MLFLOW_CLIENT.get_experiment(experiment_id=self.experiment_id).name

    # This is a property so that we can use `getattr`` with this in `generate_vertical_html_table`
    @property
    def _get_mlflow_path_html_with_link(self) -> str:
        """
        Get HTML for a link to the MLflow run page for this run.

        Only supported for TrainingRuns that have an experiment ID and run ID.
        """
        if not self.run_id or not self.experiment_id:
            raise ValueError('Run ID or Experiment ID is not set. ')
        run_name = _MLFLOW_CLIENT.get_run(run_id=self.run_id).info.run_name
        return self._get_mlflow_run_link_html(link_text=f'{self.experiment_path}/{run_name}')

    def _get_mlflow_message_html(self, num_runs: int = 1) -> str:
        """
        Get HTML for a message to display the MLflow run and experiment attributed to this run in the format:
        "Logged a run to an experiment in MLflow." with links to the run and experiment

        Also prints the experiment ID for easy cancellation/deletion.

        For older runs that do not have an experiment ID and run ID, this returns a message with the submitted
        experiment path and no links.
        """
        if self.experiment_id is None:
            raise ValueError('Experiment ID is not set')
        if not self.run_id or not self.experiment_id:
            return f'Logged to {self.submitted_experiment_path}'
        copyable_experiment_id = self._get_copyable_experiment_id_html(self.experiment_id)
        if num_runs > 1:
            return f'Logged {num_runs} runs to an {self._get_mlflow_experiment_link_html("experiment")} in MLflow. ' \
                f'Experiment ID: {copyable_experiment_id}'
        return f'Logged a {self._get_mlflow_run_link_html("run")} to an ' \
            f'{self._get_mlflow_experiment_link_html("experiment")} in MLflow. Experiment ID: {copyable_experiment_id}'

    def _get_mlflow_experiment_link_html(self, link_text: str = 'experiment') -> str:
        """
        Get a link for the MLflow experiment using the given link text.
        This link will open the MLflow experiment page in a new tab.

        Raises:
            ValueError: If the experiment ID is not set
        """
        if self.experiment_id is None:
            raise ValueError('Experiment ID is not set')
        if is_running_in_databricks_notebook():
            host = ''
        else:
            host, _ = get_host_and_token_from_env()
        path = urljoin(host, f'/ml/experiments/{self.experiment_id}')
        return f"<a href={path} class target='_blank'>{link_text}</a>"

    def _get_mlflow_run_link_html(self, link_text: str = 'run') -> str:
        """
        Get a link for the MLflow run using the given link text. This link will open the MLflow run page in a new tab.

        Raises:
            ValueError: If the run ID or experiment ID is not set
        """
        if self.run_id is None or self.experiment_id is None:
            raise ValueError('Run ID or Experiment ID is not set')
        if is_running_in_databricks_notebook():
            host = ''
        else:
            host, _ = get_host_and_token_from_env()
        path = urljoin(host, f'/ml/experiments/{self.experiment_id}/runs/{self.run_id}')
        return f"<a href={path} class target='_blank'>{link_text}</a>"

    def _get_copyable_experiment_id_html(self, experiment_id: str) -> str:
        """
        Generate HTML for a read-only input field that contains the experiment ID.
        This allows the experiment ID to be easily copied.
        """
        return f'<input type="text" value="{experiment_id}"' \
            'readonly="readonly" style="border:none;background:none;cursor:pointer;" onclick="this.select();">'



"""Interactive utilities for the KIE notebook"""
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union, cast, get_args, get_origin

import yaml
from IPython.display import clear_output, display
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from rich.prompt import Prompt

from databricks.kie.kie_schema import ModelManager, dict_to_model, model_to_dict

if TYPE_CHECKING:
    from ipywidgets import Textarea


def walkthrough_descriptions(model: Type[BaseModel],
                             current_path: Optional[str] = None,
                             current_outputs: Optional[List[str]] = None):
    """
    Interactively add descriptions to fields of an existing Pydantic model.
    Returns a new model with the added descriptions.
    """
    editor = ModelManager(model)
    model_name = editor.model.__name__

    def format_field(name: str, field: FieldInfo, prefix: str = "", with_description: bool = True) -> str:
        formatted = f"{prefix}{name} ({field.annotation.__name__})"
        if with_description and field.description:
            formatted = ": ".join([formatted, field.description])
        return formatted

    def get_full_path(name: str, current_path: Optional[str] = None) -> str:
        if not current_path:
            return name
        return " → ".join([current_path, name])

    current_outputs = current_outputs or []

    def refresh_display():
        clear_output(wait=True)
        for l in current_outputs:
            print(l)

    if not current_path:
        current_outputs.append(f"Adding descriptions for {model_name} fields:")
        current_outputs.append("(Press Enter to skip a field or keep existing description)\n")

    refresh_display()

    # Gather descriptions for each field
    for field_name, field in editor.model.model_fields.items():
        prompt = format_field(f"[bold]{get_full_path(field_name, current_path)}[/]", field, " > ", False)
        desc = Prompt.ask(prompt, default=field.description)
        #desc = input(prompt)
        # Clear the input line
        if desc and desc.strip():
            editor.edit_descriptions(**{field_name: desc.strip()})

        # Display the progress
        current_outputs.append(
            format_field(
                name=f"\033[1m{get_full_path(field_name, current_path)}\033[0m",  # Bold the name with ANSI
                field=editor.model.model_fields[field_name],  # Get the updated field
                prefix=" ✔ ",  # Prefix as completed
                with_description=True,
            ))

        refresh_display()

        # If the field is another BaseModel, recurse into it
        type_ = field.annotation
        make_optional = False
        if get_origin(type_) is Union:  # This is an optional field, strip the optional
            make_optional = True
            type_ = [t for t in get_args(type_) if t is not None][0]

        # This is an array, check the first item
        if get_origin(type_) is list:
            args = get_args(type_)
            if isinstance(args[0], type) and issubclass(args[0], BaseModel):
                # This is an array of models
                type_ = walkthrough_descriptions(args[0], field_name, current_outputs)
                type_ = List[type_]
        elif isinstance(type_, type) and issubclass(type_, BaseModel):
            # This is a nested model
            type_ = walkthrough_descriptions(type_, field_name, current_outputs)

        if make_optional:
            type_ = Optional[type_]
        editor.edit_types(**{field_name: type_})

    return editor.model


def show_schema_editor(output_schema: Type[BaseModel]) -> 'Textarea':
    """Show a YAML-based editor to quickly edit your desired output schema. 
    
    The editor updates automatically, so you don't have to confirm your outputs.
    When you are done, run:
    ```
    # handle = show_schema_editor(output_schema)
    
    output_schema = update_schema_from_handle(handle)
    ```
    """
    from ipywidgets import Textarea  # pylint: disable=import-outside-toplevel

    initial = cast(str, yaml.dump(model_to_dict(output_schema)))
    w = Textarea(value=initial,
                 description="Extract:",
                 layout={
                     "height": f"{len(initial.splitlines()) * 1.3}em",
                     "width": "100%"
                 },
                 disabled=False)
    display(w)
    return w


def update_schema_from_handle(handle: 'Textarea', existing_schema: Type[BaseModel]) -> Tuple[Type[BaseModel], bool]:
    """Update your output schema using the YAML stored in the text box referenced by `handle`. 
    
    Returns a tuple of the new schema + a boolean stating if the schema has changedd
    
    `handle` should be created by first calling:
    ```
    handle = show_schema_editor(output_schema)
    ```
    """
    existing_str = yaml.dump(model_to_dict(existing_schema))
    has_changed = existing_str != handle.value

    try:
        schema_yaml = yaml.safe_load(handle.value)
    except Exception as e:
        raise ValueError("Could not decode YAML of schema. Double check the text box or try running it again") from e
    if schema_yaml is None:
        raise ValueError("Text box is empty! Try running it again")
    schema_yaml = cast(Dict[str, Any], schema_yaml)
    return dict_to_model(schema_yaml), has_changed


def check_schema_has_descriptions(schema: Type[BaseModel]) -> bool:
    # Check the first few descriptions - this doesn't need to be exhaustive
    count = 0
    for _, field in schema.model_fields.items():
        if not field.description:
            return False
        count += 1
        if count >= 5:
            break
    return True



"""Implementation of task specific Text2Text metrics."""

from typing import Callable, List

import evaluate
from databricks.agents.evals import metric
from mlflow.evaluation import Assessment
from mlflow.metrics import MetricValue

from databricks.kie.eval_utils import PrimitiveTypes, fuzzy_match, normalized_match  # pylint: disable=ungrouped-imports
from databricks.kie.t2t_schema import InstructionType
from databricks.kie.text_utils import normalize_text

CLASSIFICATION_MATCH_NAME = "classification_match"
FUZZY_MATCH_FOR_SUBSPAN_EXTRACTION_NAME = "fuzzy_match_for_subspan_extraction"
TOKENS_OVERLAP_FOR_SUBSPAN_EXTRACTION_NAME = "tokens_overlap_for_subspan_extraction"
ROUGE_L_SCORE_FOR_SUBSPAN_EXTRACTION_NAME = "rouge_L_score_for_subspan_extraction"
ROUGE_L_SCORE_FOR_SUMMARIZATION_NAME = "rouge_L_score_for_summarization"
BLEU_SCORE_FOR_SUMMARIZATION_NAME = "bleu_score_for_summarization"

# CLASSIFICATION METRICS


def is_classification_match(request: PrimitiveTypes, prediction: PrimitiveTypes, target: PrimitiveTypes,
                            **kwargs) -> Assessment:
    """Normalized (+ exact) matching check between prediction and target."""
    if type(prediction) != type(target):  # pylint: disable=unidiomatic-typecheck
        return Assessment(name=kwargs.get("metric_name", CLASSIFICATION_MATCH_NAME),
                          value=False,
                          rationale="Prediction and target are not the same type")

    # request is unused in this metric, but needs to be passed in
    match_assessment: Assessment = normalized_match(request, prediction, target)
    return Assessment(name=kwargs.get("metric_name", CLASSIFICATION_MATCH_NAME),
                      value=match_assessment.value,
                      rationale=match_assessment.rationale)


# SUB-SPAN EXTRACTION METRICS


def fuzzy_match_score_with_target(request: PrimitiveTypes, prediction: PrimitiveTypes, target: PrimitiveTypes,
                                  **kwargs) -> Assessment:
    """Fuzzy string matching check between prediction and target."""
    if type(prediction) != type(target):  # pylint: disable=unidiomatic-typecheck
        return Assessment(name=kwargs.get("metric_name", FUZZY_MATCH_FOR_SUBSPAN_EXTRACTION_NAME),
                          value=False,
                          rationale="Prediction and target are not the same type")

    # request is unused in this metric, but needs to be passed in
    match_assessment: Assessment = fuzzy_match(request, prediction, target)
    return Assessment(name=kwargs.get("metric_name", FUZZY_MATCH_FOR_SUBSPAN_EXTRACTION_NAME),
                      value=match_assessment.value,
                      rationale=match_assessment.rationale)


def token_overlap_percentage_with_target(
        request: PrimitiveTypes,  # pylint: disable=unused-argument
        prediction: PrimitiveTypes,
        target: PrimitiveTypes,
        **kwargs) -> Assessment:
    """Token overlap check between prediction and target. Best for keyword extraction."""
    if not isinstance(prediction, str) or not isinstance(target, str):
        return Assessment(name=kwargs.get("metric_name", TOKENS_OVERLAP_FOR_SUBSPAN_EXTRACTION_NAME),
                          value=0.0,
                          rationale="Prediction and target are not strings")

    prediction_tokens = normalize_text(prediction).split()
    target_tokens = normalize_text(target).split()
    overlap = len(set(prediction_tokens) & set(target_tokens)) / len(set(prediction_tokens))
    return Assessment(name=kwargs.get("metric_name", TOKENS_OVERLAP_FOR_SUBSPAN_EXTRACTION_NAME),
                      value=int(overlap * 100),
                      rationale=f"{int(overlap * 100)}% of tokens in prediction overlap with the target")


def rouge_l_score_with_target(
        request: PrimitiveTypes,  # pylint: disable=unused-argument
        prediction: PrimitiveTypes,
        target: PrimitiveTypes,
        **kwargs) -> Assessment:
    """RougeL score check between prediction and target."""

    if not isinstance(prediction, str) or not isinstance(target, str):
        return Assessment(name=kwargs.get("metric_name", ROUGE_L_SCORE_FOR_SUBSPAN_EXTRACTION_NAME),
                          value=0.0,
                          rationale="Prediction and/or target are not strings")
    try:
        # TODO(Ricky): Try to find a more parallelizable way to calculate the Rouge L score
        rouge_calc: MetricValue = evaluate.load("rouge").compute(predictions=[prediction],
                                                                 references=[target])["rougeL"]
    except Exception as e:  # pylint: disable=broad-except
        return Assessment(name=kwargs.get("metric_name", ROUGE_L_SCORE_FOR_SUBSPAN_EXTRACTION_NAME),
                          value=None,
                          rationale=f'Error calculating Rouge L score with inputs "{prediction}" and "{target}": {e}')
    return Assessment(name=kwargs.get("metric_name", ROUGE_L_SCORE_FOR_SUBSPAN_EXTRACTION_NAME),
                      value=rouge_calc,
                      rationale=f"Rouge L score: {rouge_calc}")


# SUMMARIZATION METRICS


def bleu_score_with_target(
        request: PrimitiveTypes,  # pylint: disable=unused-argument
        prediction: PrimitiveTypes,
        target: PrimitiveTypes,
        **kwargs) -> Assessment:
    """We use BLEU to calculate the score for the prediction and target."""
    if not isinstance(prediction, str) or not isinstance(target, str):
        return Assessment(name=kwargs.get("metric_name", BLEU_SCORE_FOR_SUMMARIZATION_NAME),
                          value=0.0,
                          rationale="Prediction and/or target are not strings")
    if len(prediction.split()) < 4 or len(target.split()) < 4:
        return Assessment(
            name=kwargs.get("metric_name", BLEU_SCORE_FOR_SUMMARIZATION_NAME),
            value=0.0,
            rationale="Prediction and/or target have less than 4 tokens, which isn't enough to calculate a BLEU score")
    try:
        # TODO(Ricky): Try to find a more parallelizable way to calculate the BLEU score
        bleu_calc: MetricValue = evaluate.load("bleu").compute(predictions=[prediction], references=[target])["bleu"]
    except Exception as e:  # pylint: disable=broad-except
        return Assessment(name=kwargs.get("metric_name", BLEU_SCORE_FOR_SUMMARIZATION_NAME),
                          value=None,
                          rationale=f'Error calculating BLEU score with inputs "{prediction}" and "{target}": {e}')
    return Assessment(name=kwargs.get("metric_name", BLEU_SCORE_FOR_SUMMARIZATION_NAME),
                      value=bleu_calc,
                      rationale=f"Bleu score with prediction and target: {bleu_calc}")


# METRIC CREATION


def create_metric(fn: Callable[[PrimitiveTypes, PrimitiveTypes, PrimitiveTypes, str], float], metric_name: str):
    """Helper function to create an mlflow metric."""

    @metric(name=metric_name)
    def wrapper(request, response, expected_response):
        # Passing in the task specific metric name here since we could have the same metrics for different tasks
        return fn(request, response, expected_response, metric_name=metric_name)

    return wrapper


CLASSIC_METRICS = {
    InstructionType.classification: [create_metric(is_classification_match, CLASSIFICATION_MATCH_NAME)],
    InstructionType.subspan_extraction: [
        create_metric(fuzzy_match_score_with_target, FUZZY_MATCH_FOR_SUBSPAN_EXTRACTION_NAME),
        create_metric(token_overlap_percentage_with_target, TOKENS_OVERLAP_FOR_SUBSPAN_EXTRACTION_NAME),
        create_metric(rouge_l_score_with_target, ROUGE_L_SCORE_FOR_SUBSPAN_EXTRACTION_NAME),
    ],
    InstructionType.summarization: [
        create_metric(rouge_l_score_with_target, ROUGE_L_SCORE_FOR_SUMMARIZATION_NAME),
        create_metric(bleu_score_with_target, BLEU_SCORE_FOR_SUMMARIZATION_NAME),
    ],
}


def create_classic_metrics(task_type: InstructionType) -> List[metric]:
    """Create an mlflow metric given a task type."""
    return CLASSIC_METRICS.get(task_type, [])



"""Simple utils for working with KIE datasets"""
import json
import os
from functools import wraps
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type

import pandas as pd
from pydantic import BaseModel, ValidationError
from pyspark.sql import functions as F
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import monotonically_increasing_id, rand
from pyspark.sql.types import StringType, StructField, StructType
from pyspark.sql.window import Window

from databricks.kie.kie_evaluator import parse_json_markdown
from databricks.model_serving.types.pt_endpoint import BaseEndpoint
from databricks.model_training.api.utils import check_if_table_exists, check_table_has_columns, get_spark

VALID_EXTENSIONS = {
    '.c', '.cpp', '.cs', '.css', '.go', '.html', '.java', '.js', '.json', '.md', '.php', '.py', '.rb', '.sh', '.tex',
    '.ts', '.txt', ''
}

MIN_LABELED_TRAIN_SAMPLES = 100
MIN_DOCS_TO_TRAIN = 50

IGNORED_CACHE_ROOT = 'kie_cache_'


def cache_on_serverless(func):

    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        spark = get_spark()
        return spark.createDataFrame(result.toPandas())

    return wrapper


def get_split_from_labeled(labeled_split_df: DataFrame, split: str, request: str, expected_response: str) -> DataFrame:
    # yapf: disable
    return (labeled_split_df
            .where(F.col('split') == split)
            .withColumn('request', F.col(request))
            .withColumn('expected_response', F.col(expected_response))
    )  # yapf: enable


@cache_on_serverless
def get_split_from_unlabeled(unlabeled_split_df: DataFrame, split: str) -> DataFrame:
    # yapf: disable
    df = (unlabeled_split_df
            .where(F.col('split') == split)
            .withColumn('request', F.udf(read_document, StringType())(F.col('doc_uri')))
    )
    return df # yapf: enable

def get_all_unlabeled(unlabeled_split_df: DataFrame) -> DataFrame:
    """Get all unlabeled data from the split df """
    # yapf: disable
    df = (unlabeled_split_df
            .withColumn('request', F.udf(read_document, StringType())(F.col('doc_uri')))) # yapf: enable
    return df


def get_valid_files(directory: str) -> List[str]:
    """ Recursively get all files in the directory that match valid extensions
    """
    all_files = []
    for root, _, files in os.walk(directory):
        if os.path.basename(root).startswith(IGNORED_CACHE_ROOT):
            # This is a cache folder - skip it
            continue
        for f in files:
            if os.path.splitext(f)[1].lower() in VALID_EXTENSIONS:
                all_files.append(os.path.join(root, f))

    return all_files


def split_labeled_data(
    df: DataFrame,
    num_grounding_samples: int = 5,
    num_val_samples: int = 25,
    seed: int = 42,
) -> DataFrame:

    # Count the total sample size
    total_samples = df.count()

    # Allocate first to grounding, then val and training
    labeled_remaining = total_samples

    num_grounding_samples = min(num_grounding_samples, labeled_remaining)
    labeled_remaining -= num_grounding_samples

    # Default split to train
    df = df.withColumn('split', F.lit('train'))

    # Create random data splits
    # Order by rand so we can sample from it
    df = df.withColumn('rand', F.rand(seed=seed)).orderBy('rand')
    df = df.withColumn('row_id', F.row_number().over(Window.partitionBy("split").orderBy('rand')))

    # Split grounding and val data
    df = df.withColumn('split',
                       F.when(F.col('row_id') <= num_grounding_samples, F.lit('grounding')).otherwise(F.col('split')))
    df = df.withColumn('split',
                       F.when((F.col('row_id') > num_grounding_samples) & \
                              (F.col('row_id') <= num_grounding_samples + num_val_samples),
                              F.lit('val')).otherwise(F.col('split')))

    return df


def split_unlabeled_data(
    dataset: str,
    num_val_samples: int = 25,
    num_grounding_samples: int = 5,
    seed: int = 42,
) -> DataFrame:

    # Find all available files in the UC volume
    all_files = get_valid_files(dataset)

    if not all_files:
        raise ValueError(f"No files found in {dataset} matching accepted file types: {','.join(VALID_EXTENSIONS)}")

    # Create a dataframe with the file paths
    spark = get_spark()
    schema = StructType([StructField('doc_uri', StringType(), True)])

    # Create DataFrame with schema
    df = spark.createDataFrame([(uri,) for uri in all_files], schema=schema).orderBy('doc_uri')

    # Calculate how many unlabeled samples we have
    total_files = len(all_files)

    # Adjust sample counts to respect max percentages
    max_grounding = max(int(total_files * 0.10), 1)  # 10% max for grounding
    max_val = max(int(total_files * 0.20), 1)  # 20% max for validation

    # Calculate needed samples respecting maximums
    grounding_needed = min(num_grounding_samples, max_grounding)
    val_needed = min(num_val_samples, max_val)

    # Default split to unused
    df = df.withColumn('split', F.lit('unused'))

    # Order by rand so we can sample from it
    df = df.withColumn('rand', F.rand(seed=seed)).orderBy('rand')
    df = df.withColumn('row_id', F.row_number().over(Window.partitionBy("split").orderBy('rand')))

    # Split grounding and val data
    running_count = 0
    if grounding_needed > 0:
        df = df.withColumn('split',
                           F.when(F.col('row_id') <= grounding_needed, F.lit('grounding')).otherwise(F.col('split')))
        running_count += grounding_needed

    if val_needed > 0:
        df = df.withColumn(
            'split',
            F.when((F.col('row_id') > running_count) & (F.col('row_id') <= running_count + val_needed),
                   F.lit('val')).otherwise(F.col('split')))
        running_count += val_needed

    return df.select('doc_uri', 'split').orderBy('doc_uri')


def read_document(path: str) -> str:
    with open(path, 'r', encoding='utf-8') as f:
        return f.read()


def get_fewshot_examples(
    df: DataFrame,
    num_examples: int,
    response_format: Type[BaseModel],
    response_column: str = 'expected_response',
) -> List[Tuple[str, Dict[str, Any]]]:
    """
    Extract valid few-shot examples from the provided DataFrame.

    Args:
        df: DataFrame containing request/response pairs
        num_examples: Number of examples to extract
        response_format: Expected format for validation
        response_column: Column containing the expected response

    Returns:
        List of (request, response) tuples
    """
    examples = []
    for row in df.limit(num_examples).collect():
        try:
            response = json.loads(row[response_column])
            response_format(**response)  # Validate response
            examples.append((row['request'], response))
            if len(examples) >= num_examples:
                break
        except (json.JSONDecodeError, ValidationError):
            continue
    return examples

def filter_valid_json(df: DataFrame, column_name: str, schema: Type[BaseModel]) -> DataFrame:
    spark = get_spark()

    # Convert to pandas
    pdf = df.toPandas()

    # Validate each row
    valid_mask = pdf[column_name].apply(lambda x: _is_valid_json(x, schema))

    # Filter and convert back to Spark
    filtered_pdf = pdf[valid_mask]
    return spark.createDataFrame(filtered_pdf, df.schema)

def _is_valid_json(maybe_json_str: str, schema: Type[BaseModel]) -> bool:
    try:
        schema(**json.loads(maybe_json_str))
        return True
    except:  # pylint: disable=bare-except
        return False


def _make_training_row(
    prompt: str,
    request: str,
    response: str,
) -> dict:
    user_message = prompt + '\n\nDocument:\n\n' + request.replace('"', '').replace('“', '').replace(
        '”', '') + '\n\n' + prompt
    messages = [{
        'role': 'user',
        'content': user_message
    }, {
        'role': 'assistant',
        'content': parse_json_markdown(response)
    }]
    messages_line = {'messages': messages}
    return messages_line


def process_for_sampling(df: DataFrame, seed: int = 42) -> DataFrame:
    spark = get_spark()
    df_shuffled = df.orderBy(rand(seed=seed))
    df_with_id = df_shuffled.withColumn('id', monotonically_increasing_id())
    df_with_id_static = spark.createDataFrame(df_with_id.toPandas())
    return df_with_id_static


def batched_iterator(df: DataFrame, batch_size: int = 1000) -> Iterable[pd.DataFrame]:
    current_max_index = 0
    while current_max_index < df.count():
        yield df.where((df.id >= current_max_index) & (df.id < current_max_index + batch_size)).withColumn(
            'request',
            F.udf(read_document, StringType())(F.col('doc_uri'))).toPandas()  # type: ignore
        current_max_index += batch_size


def write_batch_to_cache(prompt: str,
                         documents: List[str],
                         responses: List[str],
                         response_format: Type[BaseModel],
                         train_jsonl_cache_path: str):
    lines = []
    for request, response in zip(documents, responses):
        if not _is_valid_json(response, response_format):
            continue

        training_row = _make_training_row(
            prompt=prompt,
            request=request,
            response=response,
        )
        lines.append(json.dumps(training_row))

    with open(train_jsonl_cache_path, 'a', encoding='utf-8') as fh:
        fh.write('\n'.join(lines) + '\n' if lines else '')



def create_training_data(
    unlabeled_split_df: DataFrame,
    prompt: str,
    response_format: Type[BaseModel],
    train_jsonl_cache_path: str,
    endpoint: BaseEndpoint,
    token_budget: int,
    min_docs_to_train: int = MIN_DOCS_TO_TRAIN,
) -> Optional[Tuple[str, int]]:

    train_df = unlabeled_split_df.where((F.col('split') == 'train') | (F.col('split') == 'unused'))

    if train_df.count() < min_docs_to_train:
        return None

    ready_for_sampling = process_for_sampling(train_df)

    token_budget_remaining = token_budget
    for batch in batched_iterator(ready_for_sampling):
        documents = batch['request'].tolist()

        responses = endpoint.generate_batch(
            prompts=documents,
            system_prompt=prompt,
            total_max_tokens=token_budget_remaining,
            response_format=response_format,
            show_progress=False,
        )

        actual_responses = [r.response for r in responses]

        write_batch_to_cache(prompt, documents, actual_responses, response_format, train_jsonl_cache_path)
        token_budget_remaining -= sum(r.total_tokens for r in responses)
    return train_jsonl_cache_path, token_budget - token_budget_remaining


def validate_row_count(
    df: DataFrame,
    required_count: int = 10,
    recommended_count: int = 1000,
) -> None:
    row_count = df.count()
    if row_count < required_count:
        raise ValueError(f'Insufficient data. Found {row_count} rows, expected at least {required_count}.')
    if row_count < recommended_count:
        print(f'Warning: Found {row_count} rows, recommended at least {recommended_count} for best results.')


def to_row_based_format(column_based_data: Dict[str, List[str]]) -> List[Dict[str, str]]:
    """
    Converts column-based data to row-based format.

    Args:
        column_based_data (dict): A dictionary where keys are column names and values are lists of column values.

    Returns:
        list: A list of dictionaries where each dictionary represents a row of data.
    """
    row_based_data = []
    num_rows = len(next(iter(column_based_data.values())))

    if len(set(len(values) for values in column_based_data.items())) != 1:
        raise ValueError('column-based data must have the same number of rows for all columns')

    for i in range(num_rows):
        row = {key: value[i] for key, value in column_based_data.items()}
        row_based_data.append(row)
    return row_based_data


def to_column_based_format(row_based_data: List[Dict[str, str]]) -> Dict[str, List[str]]:
    """
    Converts row-based data to column-based format.

    Args:
        row_based_data (list): A list of dictionaries where each dictionary represents a row of data.

    Returns:
        dict: A dictionary where keys are column names and values are lists of column values.
    """
    if not row_based_data:
        return {}

    column_based_data = {key: [] for key in row_based_data[0].keys()}
    for row in row_based_data:
        if set(row.keys()) != set(column_based_data.keys()):
            raise ValueError('all rows must have the same keys')

        for key, value in row.items():
            column_based_data[key].append(value)
    return column_based_data


def read_from_table_with_columns(table_path: str, columns: Tuple[str], count: int) -> List[Dict[str, Any]]:
    """
    Reads specified columns from a table and returns a DataFrame with a limited number of rows.

    Args:
        table_path (str): The path to the table.
        columns (List[str]): The list of columns to read.
        count (int): The number of rows to read.

    Returns:
        list: A list of dictionaries where each dictionary represents a row of the DataFrame.
    """
    if not check_if_table_exists(table_path):
        raise ValueError(f'{table_path} is not a valid table path. Please check that your table exists.')
    if not check_table_has_columns(table_path, columns):
        raise ValueError(f'{table_path} does not have expected columns: {columns}')

    spark = get_spark()
    df = spark.read.table(table_path).select(*columns).limit(count).collect()
    return [datum.asDict() for datum in df]

def read_from_jsonl(file_path: str, count: Optional[int]= None) -> List[Dict[str, Any]]:
    """
    Reads a JSONL file and returns a specified number of rows.

    Args:
        file_path (str): The path to the JSONL file.
        count (int): Optional number of rows to read.

    Returns:
        list: A list of dictionaries where each dictionary represents a row of the JSONL file.
    """
    json_obj = pd.read_json(path_or_buf=file_path, lines=True, nrows=count)
    return json_obj.to_dict(orient='records')



"""Factory for KIE schemas"""
import hashlib
import json
from pathlib import Path
from types import UnionType
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union, cast, get_args, get_origin

from pydantic import BaseModel, ConfigDict, Field, create_model
from pydantic.fields import FieldInfo

_valid_types = Union[Type, Type[Any], UnionType]

PYTHON_TO_YAML_TYPE_MAP: Dict[_valid_types, str] = {
    str: "str",
    int: "int",
    bool: "bool",
    float: "float",
    Any: "any",
}

YAML_TO_PYTHON_TYPE_MAP: Dict[str, _valid_types] = {v: k for k, v in PYTHON_TO_YAML_TYPE_MAP.items()}

JSON_TO_PYTHON_TYPE_MAP: Dict[str, _valid_types] = {
    "string": str,
    "integer": int,
    "bool": bool,
    "number": float,
    "any": Any,
}

CONFIG_DICT = ConfigDict(protected_namespaces=())


def create_custom_model(model_name: str, **field_definitions: Tuple[type, FieldInfo]) -> type[BaseModel]:
    """Create a Pydantic model with custom configuration."""
    return create_model(model_name, __config__=CONFIG_DICT, **field_definitions)  # type: ignore


class ModelFactory:
    """Factory class for creating Pydantic models from JSON examples."""

    @staticmethod
    def infer_type(value: Any, key: str = "DyanmicModel") -> type:
        """Infer Python/Pydantic type from a JSON value."""
        if isinstance(value, bool):
            return bool
        elif isinstance(value, int):
            return int
        elif isinstance(value, float):
            return float
        elif isinstance(value, str):
            return str
        elif isinstance(value, dict):
            return ModelFactory.from_dict(value, key)
        elif isinstance(value, list):
            if not value:  # Empty list
                return List[Any]
            # Infer type from first element
            element_type = ModelFactory.infer_type(value[0], key)
            return List[element_type]
        else:
            return Type[Any]

    @staticmethod
    def create_field_definitions(data: Dict[str, Any]) -> Dict[str, Tuple[type, FieldInfo]]:
        """Create field definitions for Pydantic model."""
        fields = {}
        for key, value in data.items():
            if value is None:
                # Handle optional fields
                field_type = Optional[Any]
            else:
                field_type = ModelFactory.infer_type(value, key)

            # Create field definition
            # Without knowing otherwise, we'll call every  field "Optional"
            fields[key] = (Optional[field_type], Field(description=""))

        return fields

    @staticmethod
    def create_model_from_dict(data: Dict[str, Any], model_name: str = "DynamicModel") -> type[BaseModel]:
        """Create a Pydantic model from a dictionary."""
        field_definitions = ModelFactory.create_field_definitions(data)
        # Ignore typing: field_definitions keeps matching with other kwargs that are reserved.
        # Unsure how to clean that up
        return create_custom_model(model_name, **field_definitions)

    @classmethod
    def from_dict(cls, data: Dict[str, Any], model_name: str = "DynamicModel") -> type[BaseModel]:
        """Create a model from a dictionary."""
        return cls.create_model_from_dict(data, model_name)

    @classmethod
    def from_examples(cls, examples: List[Dict[str, Any]], model_name: str = "DynamicModel") -> type[BaseModel]:
        """Build a JSON schema from a dictionary of examples."""
        # Combine all examples into one to cover as many of the fields as possible
        # Note: This assumes that all examples of a given field are of the same type
        combined_example = {key: value for example in examples for key, value in example.items()}

        return cls.from_dict(combined_example, model_name)

    @staticmethod
    def to_file(model: Type[BaseModel], path: Union[Path, str]) -> None:
        """Save a Pydantic model class schema to file."""

        path = Path(path)
        path.parent.mkdir(parents=True, exist_ok=True)

        # Get full schema with definitions for nested models
        schema = model.model_json_schema()

        with path.open('w', encoding="utf-8") as f:
            json.dump({'_model_name': model.__name__, 'schema': schema}, f, indent=2)

    @classmethod
    def from_file(cls, path: Union[Path, str]) -> Type[BaseModel]:
        """Load a Pydantic model class from a schema file."""

        path = Path(path)
        with path.open('r', encoding="utf-8") as f:
            data = json.load(f)

        schema = data['schema']

        def get_item_type(field_schema: Dict[str, Any]) -> _valid_types:
            """Get type from field schema, handling complex types like Optional[List[str]]."""

            # Handle anyOf (only allowed for Optional types)
            if 'anyOf' in field_schema:
                types = field_schema['anyOf']
                # Check if this is an Optional type (one null and one non-null)
                if len(types) == 2 and any(t.get('type') == 'null' for t in types):
                    # Get the non-null type - e.g. {'type': 'string'}
                    non_null_type = next(t for t in types if t.get('type') != 'null')
                    found_type = get_item_type(non_null_type)
                    return Optional[found_type]
                else:
                    # Raise an error for other Union types
                    raise TypeError("Union types are only allowed for Optional fields."
                                    f"Found schema: {field_schema}")

            # Handle direct type definitions
            if 'type' in field_schema:
                found_type = field_schema['type']

                # Handle arrays
                if found_type == 'array' and 'items' in field_schema:
                    if '$ref' in field_schema['items']:
                        ref_name = field_schema['items']['$ref'].split('/')[-1]
                        ref_schema = schema.get('$defs', {}).get(ref_name, {})
                        item_type = create_model_from_schema(ref_schema, ref_name)
                    else:
                        item_type = get_item_type(field_schema['items'])
                    return List[item_type]

                # Handle basic types
                return JSON_TO_PYTHON_TYPE_MAP.get(found_type, Any)

            return Any

        def create_model_from_schema(schema_data: Dict[str, Any], model_name: str) -> Type[BaseModel]:
            fields = {}

            for field_name, field_schema in schema_data.get('properties', {}).items():
                # Handle references to nested models
                if '$ref' in field_schema:
                    ref_name = field_schema['$ref'].split('/')[-1]
                    ref_schema = schema.get('$defs', {}).get(ref_name, {})
                    field_type = create_model_from_schema(ref_schema, ref_name)
                else:
                    field_type = get_item_type(field_schema)

                # Create field with description and title if present
                field_info = {}
                if 'description' in field_schema:
                    field_info['description'] = field_schema['description']
                if 'title' in field_schema:
                    field_info['title'] = field_schema['title']

                # Setup the type, Field tuple
                # Note all fields are considered "required", but some are nullable.
                fields[field_name] = (field_type, Field(**field_info))

            return create_custom_model(model_name, **fields)

        return create_model_from_schema(schema, data['_model_name'])


T = TypeVar('T', bound=BaseModel)


class ModelManager:
    """Manager for editing fields on a Pydantic model"""

    def __init__(self,
                 model: Type[T],
                 parent: Optional['ModelManager'] = None,
                 field_name: Optional[str] = None) -> None:
        self.model = model
        self.parent = parent
        self.field_name = field_name

        if (self.parent is None) is not (self.field_name is None):
            raise ValueError("Must specify both `parent` and `field_name` if one of them is specified")

    def __getitem__(self, field_name: str) -> 'ModelManager':
        if field_name not in self.model.model_fields:
            raise KeyError(f"Field {field_name} not found")

        field = self.model.model_fields[field_name]
        if not (isinstance(field.annotation, type) and issubclass(field.annotation, BaseModel)):
            raise TypeError(f"Field {field_name} is not a nested model")

        return ModelManager(field.annotation, self, field_name)

    @staticmethod
    def _copy_field(f: FieldInfo, **updates) -> Tuple[type, FieldInfo]:
        annotation = updates.pop("annotation", f.annotation)
        copied = {
            "description": f.description,
            "title": f.title,
            "default": f.default,
            "json_schema_extra": f.json_schema_extra,
        }
        copied.update(**updates)
        return (annotation, Field(**copied))

    def _update_model_with_fields(self, **fields: Tuple[type, FieldInfo]) -> Type[BaseModel]:
        new_model = create_custom_model(f"{self.model.__name__}", **fields)  # type: ignore
        self.model = cast(Type[T], new_model)
        return self.model

    def _update_parents(self) -> Type[BaseModel]:
        if not self.parent:
            return self.model

        current = self
        while current.parent is not None:
            assert current.field_name is not None
            current.parent.edit_types(**{current.field_name: current.model})
            current = current.parent
        return current.model

    def add_field(self, field_name: str, field_type: type, field_description: str = "") -> type[BaseModel]:

        # Add existing fields
        fields = {name: self._copy_field(field) for name, field in self.model.model_fields.items()}

        # Add new field
        fields[field_name] = (field_type, Field(description=field_description))

        self._update_model_with_fields(**fields)

        return self._update_parents()

    def edit_types(self, **type_updates: type) -> Type[BaseModel]:

        fields = {
            name: self._copy_field(field, annotation=type_updates.get(name, field.annotation))
            for name, field in self.model.model_fields.items()
        }
        self._update_model_with_fields(**fields)

        return self._update_parents()

    def edit_descriptions(self, **updates) -> Type[BaseModel]:

        fields = {
            name: self._copy_field(field, description=updates.get(name, field.description))
            for name, field in self.model.model_fields.items()
        }
        self._update_model_with_fields(**fields)

        return self._update_parents()


def model_to_dict(model: Type[BaseModel]) -> Dict[str, Any]:
    """Convert a Pydantic model to a dictionary describing its structure.

    Args:
        model: A Pydantic model class
        
    Returns:
        Dict with field names as keys and dicts containing:
            - type: String for basic types, dict for nested models, list for arrays 
            - description: Field description string
            
    Example:
        class User(BaseModel):
            name: str = Field(description="User name")
            tags: List[str]
            
        # Returns:
        {
            'name': {'type': 'str', 'description': 'User name'},
            'tags': {'type': [{'type': 'str'}], 'description': ''}
        }
    """
    output = {}
    for key, field in model.model_fields.items():
        type_ = field.annotation
        value: Dict[str, Any] = {"description": field.description or "", "required": True}
        if get_origin(type_) is Union:  # This is an optional field
            type_ = [t for t in get_args(type_) if t is not None][0]
            value["required"] = False

        if get_origin(type_) is list:
            args = get_args(type_)
            if isinstance(args[0], type) and issubclass(args[0], BaseModel):
                # This is an array of models
                type_value = [model_to_dict(args[0])]
            else:
                type_value = [{"type": PYTHON_TO_YAML_TYPE_MAP.get(args[0], "any")}]
        elif isinstance(type_, type) and issubclass(type_, BaseModel):
            # This is a nested model
            type_value = model_to_dict(type_)
        else:
            type_value = PYTHON_TO_YAML_TYPE_MAP.get(type_, "any")
        value['type'] = type_value
        output[key] = value
    return output


def dict_to_model(model_dict: Dict[str, Any]) -> Type[BaseModel]:
    """Create a Pydantic model from a dictionary describing its structure.

    Args:
        model_dict: Dict with field names as keys and field info dicts containing:
            - type: String for basic types, dict for nested models, list for arrays
            - description: Optional field description string
            
    Returns:
        A generated Pydantic model class with the specified structure
        
    Example:
        schema = {
            'name': {'type': 'str', 'description': 'User name'},
            'tags': {'type': [{'type': 'str'}]}
        }
        UserModel = dict_to_model(schema)
    """
    fields = {}
    for key, value in model_dict.items():
        description = value.get('description', '')
        required = value.get('required', True)

        requested_type = value.get('type')
        if not requested_type:
            raise ValueError(f"Field {key} is missing a type, please specify one in your schema.")

        type_ = Any
        if isinstance(requested_type, list):
            item = requested_type[0]
            if 'type' in item:
                requested_inner_type = item['type']
                if requested_inner_type not in YAML_TO_PYTHON_TYPE_MAP:
                    raise ValueError(f"Array field {key} has an unknown type: {requested_inner_type}. " +
                                     "Please update your schema to provide a valid type.")
                inner_type = YAML_TO_PYTHON_TYPE_MAP.get(requested_inner_type, Any)
                if inner_type == Any:
                    raise ValueError(f"Array field {key} has a type of any, which is not supported. " +
                                     "Please update your schema to provide a specific type.")

                type_ = List[inner_type]
            else:
                item_type = dict_to_model(item)
                type_ = List[item_type]
        elif isinstance(requested_type, dict):
            # Nested model
            type_ = dict_to_model(requested_type)
        elif requested_type in YAML_TO_PYTHON_TYPE_MAP:
            # Basic type
            type_ = YAML_TO_PYTHON_TYPE_MAP.get(requested_type)
        else:
            raise ValueError(
                f"Field {key} has an unknown type: {requested_type}. Please update your schema to provide a valid type."
            )

        if type_ == Any:
            raise ValueError(f"Field {key} has a type of any, which is not supported. " +
                             "Please update your schema to provide a specific type.")

        if not required:
            type_ = Optional[type_]
        fields[key] = (type_, Field(description=description))

    return create_custom_model("Model", **fields)


def get_schema_hash(response_format: Type[BaseModel]) -> str:
    schema = json.dumps(response_format.model_json_schema())
    return hashlib.md5(schema.encode()).hexdigest()[:8]



"""
Engine code for the GenAI API.
"""
from .engine import (get_return_response, retry_with_backoff, run_paginated_mapi_request, run_plural_mapi_request,
                     run_singular_mapi_request)

__all__ = [
    'get_return_response', 'retry_with_backoff', 'run_paginated_mapi_request', 'run_plural_mapi_request',
    'run_singular_mapi_request'
]



""" Common models
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Dict, Generic, Iterator, List, TypeVar

O = TypeVar('O', bound=type(dataclass))


class ObjectType(Enum):
    """ Enum for Types of Objects Allowed """

    TRAINING_RUN = 'training_run'
    RUN_EVENT = 'run_event'
    TRAINING_MODEL_CONSTANTS = 'training_model_constants'
    TRAINING_CONSTANTS = 'training_constants'

    UNKNOWN = 'unknown'

    def _get_display_columns(self) -> Dict[str, str]:
        """
        This is currently used only for html display (inside a notebook)

        Ideally the CLI & notebook display will be unified

        Returns:
            Dict[str, str]: Mapping of class column name to display name
        """

        if self == ObjectType.TRAINING_RUN:
            display_columns = {
                'name': 'Name',
                'run_progress': 'Status',
                '_get_mlflow_path_html_with_link': 'MLflow Run',
                'model': 'Model',
                'learning_rate': 'Learning Rate',
                'training_duration': 'Training Duration',
                'train_data_path': 'Train Data Path',
                'register_to': 'Register To',
            }
            return display_columns

        if self == ObjectType.RUN_EVENT:
            return {
                'type': 'Type',
                'time': 'Time',
                'message': 'Message',
            }

        if self == ObjectType.TRAINING_MODEL_CONSTANTS:
            return {'display_name': 'Model Name', 'name': 'Input Name', 'max_context_length': 'Maximum Context Length'}

        return {}

    @classmethod
    def from_model_type(cls, model) -> ObjectType:
        # pylint: disable-next=import-outside-toplevel
        from databricks.model_training.types.training_constants import TrainingConstants, TrainingModelConstants
        # pylint: disable-next=import-outside-toplevel
        from databricks.model_training.types.training_run import TrainingEvent, TrainingRun
        if model == TrainingRun:
            return ObjectType.TRAINING_RUN
        if model == TrainingEvent:
            return ObjectType.RUN_EVENT
        if model == TrainingModelConstants:
            return ObjectType.TRAINING_MODEL_CONSTANTS
        if model == TrainingConstants:
            return ObjectType.TRAINING_CONSTANTS
        return ObjectType.UNKNOWN


def generate_html_table(data: List[O], attr_to_label: Dict[str, str]) -> str:
    """
    Parameters
    ----------
    data : List[O]
        List of objects
    attr_to_label : dict
        Mapping of attributes (of the object) to labels/table headers

    Returns
    -------
    str
        An HTML table as a string. The top row will be a header consisting
        of the provided labels, and every other row will correspond to the
        provided objects
    """
    res = []
    res.append("<table border=\"1\" class=\"dataframe\">")

    # header
    res.append('<thead>')
    res.append("<tr style=\"text-align: right;\">")
    for col in attr_to_label.values():
        res.append(f'<th>{col}</th>')
    res.append('</tr>')
    res.append('</thead>')

    # body
    res.append('<tbody>')
    for row in data:
        res.append('<tr>')
        for col in attr_to_label:
            value = getattr(row, col, '')
            res.append(f"<td>{value if value else '-'}</td>")
        res.append('</tr>')
    res.append('</tbody>')

    res.append('</table>')
    return '\n'.join(res)


def generate_vertical_html_table(data: List[O], attr_to_label: Dict[str, str]) -> str:
    """
    Parameters
    ----------
    data : List[O]
        List of objects
    attr_to_label : dict
        Mapping of attributes (of the object) to labels/table headers

    Returns
    -------
    str
        An HTML table as a string. The leftmost column will be the headers provided
        as attr labels. The remaining columns will be the attributes of the objects.
    """
    res = []
    res.append("<table border='1' class='dataframe'>")
    res.append('<tbody>')
    for attr, label in attr_to_label.items():
        res.append('<tr>')
        res.append(f'<th>{label}</th>')
        for obj in data:
            value = getattr(obj, attr, None)
            res.append(f'<td>{value}</td>')
        res.append('</tr>')
    res.append('</tbody>\n</table>')
    return '\n'.join(res)


class ObjectList(Generic[O]):
    """Common helper for list of objects
    """

    def __init__(self, data: List[O], obj_type: ObjectType):
        self.data = data
        self.type = obj_type

    def __repr__(self) -> str:
        return f'List{self.data}'

    def __iter__(self) -> Iterator[O]:
        return iter(self.data)

    def __getitem__(self, index: int):
        return self.data[index]

    def __setitem__(self, index, value):
        self.data[index] = value

    def insert(self, index, item):
        self.data.insert(index, item)

    def append(self, item):
        self.data.append(item)

    def extend(self, item):
        self.data.extend(item)

    def __len__(self) -> int:
        return len(self.data)

    @property
    def display_columns(self) -> Dict[str, str]:
        return self.type._get_display_columns()  # pylint: disable=protected-access

    def __str__(self):
        return '[' + ', '.join([str(i) for i in self.data]) + ']'

    def _repr_html_(self) -> str:
        return generate_html_table(self.data, self.display_columns)

    def to_pandas(self):
        try:
            # pylint: disable=import-outside-toplevel
            import pandas as pd  # type: ignore
        except ImportError as e:
            raise ImportError('Please install pandas to use this feature') from e

        cols = self.display_columns
        res = {col: [] for col in cols}
        for row in self.data:
            for col in cols:
                value = getattr(row, col)
                res[col].append(value if value else '')

        return pd.DataFrame(data=res)



"""Utils for model training API"""
import json
import logging
import os
import re
import uuid
from datetime import datetime
from enum import Enum
from typing import Any, List, Optional

from databricks.sdk import WorkspaceClient
from databricks.sdk import errors as dbx_errors
from databricks.sdk.service.compute import ClusterDetails
from datasets import get_dataset_split_names
from mlflow import MlflowClient
from packaging import version

# pylint: disable=ungrouped-imports
from databricks.model_training.api.exceptions import ValidationError
from databricks.model_training.api.utils import (_DATABRICKS_CONFIG_PROFILE_ENV, _LOCAL_DEV_CONFIG_PROFILE,
                                                 _TEST_CONFIG_PROFILE, get_spark)
from databricks.model_training.types.train_config import TrainTaskType

logger = logging.getLogger(__name__)

_UC_VOLUME_LIST_API_ENDPOINT = '/api/2.0/fs/list'
_UC_VOLUME_FILES_API_ENDPOINT = '/api/2.0/fs/files'
_UC_PERIMISSIONS_API_ENDPOINT = '/api/2.1/unity-catalog/effective-permissions'
MIN_DBR_VERSION = version.parse('12.2')
DB_CONNECT_DBR_VERSION = version.parse('14.1')
MINIMUM_DB_CONNECT_DBR_VERSION = '14.1'
MINIMUM_SQ_CONNECT_DBR_VERSION = '12.2'
SAVE_FOLDER_PATH = 'dbfs:/databricks/mlflow-tracking/{mlflow_experiment_id}/{mlflow_run_id}/artifacts'


class SupportedDataFormats(Enum):
    UC_VOLUME = 'UC_VOLUME'
    HF_DATASET = 'HF_DATASET'
    DELTA_TABLE = 'DELTA_TABLE'


def validate_register_to(path: str) -> None:
    """
    Validates that the register_to input is one of three inputs
    1. catalog.schema
    2. catalog.schema.existing_model_name
    3. catalog.schema.new_model_name

    For 1 and 3, we must check that a user can create a model in the schema.
    For 2, we must see that the user can update the existing model.
    """
    split_path: List[str] = path.split('.')
    model_name = None
    if len(split_path) == 2:
        catalog, schema_name = split_path
    elif len(split_path) == 3:
        catalog, schema_name, model_name = split_path
    else:
        raise ValidationError(f'register_to must be in the format '
                              f'catalog.schema or catalog.schema.model_name, but got {path}')
    for component in split_path:
        if len(component) == 0:
            raise ValidationError(f'register_to must be in the format '
                                  f'catalog.schema or catalog.schema.model_name, but got {path}')
    client = WorkspaceClient()
    if model_name is not None:
        try:
            model_info = client.registered_models.get(path)
        except dbx_errors.NotFound:
            # we will create the model in the next step
            pass
        except dbx_errors.PermissionDenied as e:
            raise ValidationError(f'User does not have permission to access model "{path}" for register_to.') from e
        except dbx_errors.DatabricksError:
            # we except this just in case the ResourceDoesNotExist error changes
            # assume this means that the user needs to create the model
            pass
        else:
            # User can access model, make sure that they would theoretically be able to make a new version.
            # Note: to create the new version, we would have to use the MLflow API and submit both a run_id
            # and model artifacts. This is too involved, so we will instead just check ownership and permissions.
            # See PR #45 for more details. TODO: replace with a dry run call once this is implemented
            if model_info.owner != client.current_user.me().user_name:
                raise ValidationError(f'User must be the owner of the model {path} to register to it.')
            validate_uc_permissions(catalog,
                                    'catalog', ['ALL_PRIVILEGES', 'USE_CATALOG'],
                                    input_name='register_to',
                                    client=client)
            validate_uc_permissions(f'{catalog}.{schema_name}',
                                    'schema', ['ALL_PRIVILEGES', 'USE_SCHEMA'],
                                    input_name='register_to',
                                    client=client)
            return
    else:
        model_name = f'databricks-training-credentials-test-{uuid.uuid4()}'
    try:
        # TODO: replace with a dry run call once this is implemented
        model_attempt = client.registered_models.create(catalog_name=catalog, schema_name=schema_name, name=model_name)
        # if create goes through, delete should be safe (delete perms are a subset of create, and the user is the owner
        # since they just created the model)
        if model_attempt.full_name is None:
            raise ValidationError(f'Failed to create model for register_to at {path}.')
        client.registered_models.delete(model_attempt.full_name)
    except dbx_errors.NotFound as e:
        # thrown in the case that at least one of catalog or schema is invalid
        raise ValidationError(f'Could not find UC catalog "{catalog}" or schema "{schema_name}" for register_to '
                              'input.') from e
    except dbx_errors.PermissionDenied as e:
        raise ValidationError(f'User does not have permission to register a model at {path}.') from e
    except Exception as e:
        raise ValidationError(f'Failed to create model for register_to at {path}.') from e


def validate_uc_path_write_permissions(uc_path: str) -> None:
    """
    Validates write permissions to the path by trying to write a test file. If it succeeds, the file will be deleted
    """
    try:
        with open(os.path.join(uc_path, ".permissions_check"), "w", encoding="utf-8") as fh:
            fh.write("test")
        os.remove(os.path.join(uc_path, ".permissions_check"))
    except OSError as e:
        raise ValidationError(f'User does not have permissions to write to Unity Catalog path {uc_path}.') from e


def validate_uc_permissions(uc_path: str,
                            uc_type: str,
                            one_of_permissions: List[str],
                            input_name: str,
                            client: Optional[WorkspaceClient] = None):
    # user must have at least ONE of the permissions in the given list
    # https://docs.databricks.com/api/workspace/grants/geteffective
    final_client: WorkspaceClient = client or WorkspaceClient()
    try:
        resp: Any = final_client.api_client.do(method='GET',
                                               path=f'{_UC_PERIMISSIONS_API_ENDPOINT}/{uc_type}/{uc_path}',
                                               query={'principal': final_client.current_user.me().user_name},
                                               headers={'Source': 'mosaicml/foundation_model_training'})
        if 'privilege_assignments' not in resp:
            raise ValidationError(f'Failed to check permissions on {uc_type} {uc_path} for input {input_name}.')
        for assignment in resp['privilege_assignments']:
            for privilege in assignment['privileges']:
                privilege_type = privilege['privilege']
                if privilege_type in one_of_permissions:
                    return
    except dbx_errors.NotFound as e:
        raise ValidationError(f'Could not find {uc_type} {uc_path} for input {input_name}.') from e
    except dbx_errors.PermissionDenied as e:
        raise ValidationError(f'User does not have permission to access {uc_type} {uc_path} for '
                              f'input {input_name}.') from e
    except Exception as a:
        raise ValidationError(f'Failed to check permissions on {uc_type} {uc_path} for input {input_name}.') from a
    raise ValidationError(
        f'User does not have one of the necessary permissions ({one_of_permissions}) on {uc_type} {uc_path} for '
        f'input {input_name}.')


def validate_delta_table(path: str, input_type: str = 'train_data_path') -> None:
    split_path = path.split('.')
    if len(split_path) != 3:
        raise ValidationError(f'Delta table input to {input_type} must be in the format '
                              f'catalog.schema.table, but got {path}.')
    for component in split_path:
        if len(component) == 0:
            raise ValidationError(f'Delta table input to {input_type} must be in the format '
                                  f'catalog.schema.table, but got {path}.')
    client = WorkspaceClient()
    try:
        client.tables.get(path)
    except dbx_errors.NotFound as e:
        raise ValidationError(f'Could not find table "{path}" for input "{input_type}"') from e
    except dbx_errors.PermissionDenied as e:
        raise ValidationError(f'User does not have permission to access table "{path}" for input "{input_type}"') from e
    except Exception as e:
        raise ValidationError(f'Failed to access table "{path}" for input "{input_type}".') from e


def validate_experiment_path(experiment_path: str) -> None:
    try:
        client = MlflowClient(tracking_uri='databricks')
        experiment = client.get_experiment_by_name(experiment_path)
        if not experiment:
            client.create_experiment(experiment_path)
    except Exception as e:
        raise ValidationError(f'Failed to get or create MLflow experiment {experiment_path}. Please make sure '
                              'your experiment path is valid.') from e


def find_a_txt_file(object_path: str) -> bool:
    # comes from Composer UCObjectStore
    client = WorkspaceClient()

    try:
        resp = client.api_client.do(method='GET',
                                    path=_UC_VOLUME_LIST_API_ENDPOINT,
                                    data=json.dumps({'path': object_path}),
                                    headers={'Source': 'mosaicml/foundation_model_training'})
    except Exception as exc:
        raise ValidationError(
            f'Failed to access Unity Catalog path {object_path}. Ensure continued pretrain input is a Unity '
            'Catalog volume path to a folder.') from exc

    # repeat GET on original path to avoid duplicate code
    stack = [object_path]

    while len(stack) > 0:
        current_path = stack.pop()

        # Note: Databricks SDK handles HTTP errors and retries.
        # See https://github.com/databricks/databricks-sdk-py/blob/v0.18.0/databricks/sdk/core.py#L125 and
        # https://github.com/databricks/databricks-sdk-py/blob/v0.18.0/databricks/sdk/retries.py#L33 .
        resp = client.api_client.do(method='GET',
                                    path=_UC_VOLUME_LIST_API_ENDPOINT,
                                    data=json.dumps({'path': current_path}),
                                    headers={'Source': 'mosaicml/foundation_model_training'})

        assert isinstance(resp, dict), 'Response is not a dictionary'

        for f in resp.get('files', []):
            fpath = f['path']
            if f['is_dir']:
                stack.append(fpath)
            else:
                if f['path'].endswith('.txt'):
                    return True
    return False


def validate_uc_path(uc_path: str, task_type: TrainTaskType) -> None:
    """
    Validates the user's read access to a Unity Catalog path. If `task_type!=CONTINUED_PRETRAIN`, ensures
    that the path ends with a jsonl file. Else, ensures that the path is a folder that contains a txt file.
    """
    if not uc_path.startswith('dbfs:/Volumes'):
        raise ValidationError('Databricks Unity Catalog Volumes paths should start with "dbfs:/Volumes".')
    path = os.path.normpath(uc_path[len('dbfs:/'):])
    dirs = path.split(os.sep)
    if len(dirs) < 4:
        raise ValidationError(f'Databricks Unity Catalog Volumes path expected to start with ' \
            f'`dbfs:/Volumes/<catalog-name>/<schema-name>/<volume-name>`. Found path={uc_path}')
    object_path = '/' + path
    if task_type == TrainTaskType.CONTINUED_PRETRAIN:
        if not find_a_txt_file(object_path):
            raise ValidationError(
                f'Could not find a .txt file in Unity Catalog path {uc_path}. Continued pretrain input must be a '
                'folder containing a .txt file.')
    else:
        if not object_path.endswith('.jsonl'):
            raise ValidationError(f'Unity Catalog input for instruction finetuning must be a jsonl file. Got {uc_path}')
        try:
            client = WorkspaceClient()
            client.api_client.do(method='HEAD', path=os.path.join(_UC_VOLUME_FILES_API_ENDPOINT, path))
        except Exception as e:
            raise ValidationError(f'Failed to access Unity Catalog path {uc_path}.') from e


def validate_hf_dataset(dataset_name_with_split: str) -> None:
    print(f'Assuming {dataset_name_with_split} is a Hugging Face dataset (not in format `dbfs:/Volumes` or '
          '`/Volumes`). Validating...')
    split_dataset_name = dataset_name_with_split.split('/')
    if len(split_dataset_name) < 2:
        raise ValidationError(
            f'Hugging Face dataset {dataset_name_with_split} must be in the format <dataset>/<split> or '
            '<entity>/<dataset>/<split>.')
    dataset_name, split = '/'.join(split_dataset_name[0:-1]), split_dataset_name[-1]
    try:
        splits = get_dataset_split_names(dataset_name)
    except Exception as e:
        raise ValidationError(
            f'Failed to access Hugging Face dataset {dataset_name_with_split}. Please make sure that the split '
            'is valid and that your dataset does not have subsets.') from e
    if split not in splits:
        raise ValidationError(f'Failed to access Hugging Face dataset {dataset_name_with_split}. Split not found.')
    print('Hugging Face dataset validation successful.')


def validate_data_prep(data_prep_cluster: Optional[str] = None):
    if data_prep_cluster is None:
        raise ValidationError(
            'Providing a delta table for foundation_model data or eval data requires specifying a data_prep_cluster_id.'
        )
    if data_prep_cluster == 'serverless':
        raise ValidationError('Serverless data preparation is not supported. Please provide a cluster ID.')
    w = WorkspaceClient()
    try:
        # Check whether user has access to cluster, replicated from foundry
        res = w.clusters.get(cluster_id=data_prep_cluster)
        data_security_mode = str(res.data_security_mode,).upper()[len('DATASECURITYMODE.'):]

        # NONE stands for No Isolation Shared
        if data_security_mode == 'NONE' or res.data_security_mode is None:
            raise ValidationError(
                f'The cluster you have provided: {data_prep_cluster} does not have data governance enabled. '
                'Please use a cluster with a data security mode other than NONE.',)
    except ValidationError as e:
        raise e
    except Exception as e:
        raise ValidationError(
            f'You do not have access to the cluster you provided: {data_prep_cluster}. Please try again with another '
            'cluster.') from e


def validate_custom_weights_path(custom_weights_path: str):
    validation_error_message = ('Custom weights path must be in the format [dbfs:]/databricks/mlflow-tracking/'
                                '<experiment_id>/<run_id>/artifacts/<path>.'
                                f'Found {custom_weights_path}')
    mlflow_dbfs_path_prefix = 'dbfs:/databricks/mlflow-tracking/'
    # dbfs will be prepended before this if the user input `/databricks`
    mlflow_custom_weights_regex = (r'^dbfs:\/databricks\/mlflow-tracking'
                                   r'\/[0-9]+\/[0-9a-z]+\/artifacts($|\/[\/a-zA-Z0-9 ()_\\\-.]*$)')
    if not re.match(mlflow_custom_weights_regex, custom_weights_path):
        raise ValidationError(validation_error_message)

    subpath = custom_weights_path[len(mlflow_dbfs_path_prefix):]
    components = subpath.split('/', maxsplit=3)

    if len(components) < 4:
        raise ValidationError(validation_error_message)

    _, run_id, _, artifact_path = components
    client = MlflowClient(tracking_uri='databricks')
    is_file = re.search(r'\.[a-zA-Z0-9]+$', artifact_path) is not None

    # check if artifact is a valid .pt or .symlink file
    if is_file:
        if not artifact_path.endswith(('.symlink', '.pt')):
            raise ValidationError(f'Provided custom_weights_path {custom_weights_path} file does not end in .symlink'
                                  'Please refer to the documentation for the custom weights path format.')

        checkpoint_dir = os.path.dirname(artifact_path)
        if any(artifact.path == artifact_path for artifact in client.list_artifacts(run_id, checkpoint_dir)):
            return
        raise ValidationError(
            f'Could not find custom_weights_path {custom_weights_path}. Please double check your path exists.')

    # check if there is a distcp file in the artifact path dir
    if any(artifact.path.endswith('.distcp') for artifact in client.list_artifacts(run_id, artifact_path)):
        return

    raise ValidationError(
        f'Could not find .distcp files in the provided custom_weights_path {custom_weights_path} folder. '
        'Please refer to the documentation for the custom weights path format.')


def is_cluster_sql(cluster_id: str) -> bool:
    # Returns True if DBR version < 14.1 and requires SqlConnect
    # Returns False if DBR version >= 14.1 and can use DBConnect
    # In dev local and testing modes, returns False.
    if os.environ.get(_DATABRICKS_CONFIG_PROFILE_ENV, '') in [_LOCAL_DEV_CONFIG_PROFILE, _TEST_CONFIG_PROFILE]:
        return False
    if cluster_id == 'serverless':
        return False
    w = WorkspaceClient()
    cluster: ClusterDetails = w.clusters.get(cluster_id=cluster_id)
    if cluster.spark_version is None or not isinstance(cluster.spark_version, str):
        raise ValidationError(
            f'The cluster you provided is not compatible: please use a cluster with a DBR version > {MIN_DBR_VERSION}')
    stripped_runtime = re.sub(r'[a-zA-Z]', '', cluster.spark_version.split('-scala')[0].replace('x-snapshot', ''))
    runtime_version = re.sub(r'[.-]*$', '', stripped_runtime)
    if version.parse(runtime_version) < MIN_DBR_VERSION:
        raise ValidationError(
            'The cluster you provided is not compatible: please use a cluster with a DBR version > {MIN_DBR_VERSION}')
    if version.parse(runtime_version) < DB_CONNECT_DBR_VERSION:
        return True
    return False


def validate_create_training_run_inputs(train_data_path: str,
                                        register_to: Optional[str] = None,
                                        experiment_path: Optional[str] = None,
                                        eval_data_path: Optional[str] = None,
                                        data_prep_cluster: Optional[str] = None,
                                        custom_weights_path: Optional[str] = None,
                                        task_type: TrainTaskType = TrainTaskType.INSTRUCTION_FINETUNE) -> None:
    validate_path_and_data_prep_cluster(train_data_path, task_type, 'train_data_path', data_prep_cluster)
    if register_to:
        validate_register_to(register_to)
    if experiment_path:
        validate_experiment_path(experiment_path)
    if eval_data_path is not None:
        validate_path_and_data_prep_cluster(eval_data_path, task_type, 'eval_data_path', data_prep_cluster)
    if custom_weights_path:
        validate_custom_weights_path(custom_weights_path)


def format_path(path: str) -> str:
    """
    Prepends `dbfs:` in front of paths that start with `/Volumes` or `/databricks`.
    """
    if isinstance(path, str) and (path.startswith('/Volumes') or path.startswith('/databricks')):
        return f'dbfs:{path}'
    else:
        return path


def validate_path_and_data_prep_cluster(data_path: str, task_type: TrainTaskType, data_path_type: str,
                                        data_prep_cluster: Optional[str]) -> None:
    """
    Validates the given data path and data prep cluster for the given task type.
    """
    data_format = validate_path(data_path, task_type, data_path_type)
    if data_format == SupportedDataFormats.DELTA_TABLE:
        validate_data_prep(data_prep_cluster)


def validate_path(data_path: str, task_type: TrainTaskType, data_path_type: str) -> SupportedDataFormats:
    """
    Validates the given data path in UC volume format, HF dataset format, or Delta table format.

    Args:
        data_path: The data path to validate.
        task_type: The training run task type.
        data_path_type: The type of data path, either 'train_data_path' or 'eval_data_path' for error messages.

    Returns:
        The data format of the path.
    """
    data_path = format_path(data_path)
    if task_type == TrainTaskType.CONTINUED_PRETRAIN:
        validate_uc_path(data_path, task_type)
        return SupportedDataFormats.UC_VOLUME
    else:  # 'INSTRUCTION_FINETUNE' or 'CHAT_COMPLETION' tasks
        if data_path.startswith('dbfs:/'):
            validate_uc_path(data_path, task_type)
            return SupportedDataFormats.UC_VOLUME
        elif '/' in data_path:  # assume HF dataset TODO: state this assumption in docs
            validate_hf_dataset(data_path)
            return SupportedDataFormats.HF_DATASET
        else:
            # if we have delta table input, we must also ensure the cluster input is valid
            validate_delta_table(data_path, data_path_type)
            return SupportedDataFormats.DELTA_TABLE


def _compare_versions(v1: str, v2: str) -> int:
    # Clean up versions to standard format
    v1_clean = v1.split('-')[0].replace('.x', '.0')
    v2_clean = v2.split('-')[0].replace('.x', '.0')

    # Parse and compare
    ver1 = version.parse(v1_clean)
    ver2 = version.parse(v2_clean)

    return (ver1 > ver2) - (ver1 < ver2)


def validate_current_dbr_version(min_version: str) -> None:
    # This is super hacky but I haven't found a better way
    spark = get_spark()
    dbr_version = spark.sql("SELECT current_version().dbr_version as version").collect()[0]["version"]
    if not _compare_versions(dbr_version, min_version) >= 0:
        raise RuntimeError(
            f"Current DBR version ({dbr_version}) is too low. Please use a cluster with DBR version >= {min_version}.")


def parse_datetime(value: str) -> str:
    """Parse a datetime string into an ISO format datetime string.

    value:
        value: Datetime string

    Returns:
        String of datetime in ISO format

    Raises:
        ValidationError: If arg is not a valid datetime format"""
    for fmt in [
            '%Y-%m-%d', '%m-%d-%Y', '%H:%M:%S.%f', '%H:%M:%S', '%Y-%m-%d %H:%M:%S.%f', '%m-%d-%Y %H:%M:%S.%f',
            '%Y-%m-%d %H:%M:%S', '%m-%d-%Y %H:%M:%S'
    ]:
        try:
            return datetime.strptime(value, fmt).astimezone().isoformat()
        except ValueError:
            pass

    raise ValidationError(
        'Invalid datetime format passed. Examples: \'2023-01-13\', \'01-12-2023, 5:32:23.34\', \'2023-12-30 05:34:23\'')


def validate_table_does_not_exist(table_name: str):
    """Validates that the table does not already exist in the workspace.
    
    Args:
        table_name: The name of the table to validate.
    
    Raises:
        ValueError: If the table already exists.
    """
    w = WorkspaceClient()
    r = w.tables.exists(table_name)
    if r.table_exists:
        raise ValidationError(f"Table {table_name} already exists. Please provide a different table name.")



"""Databricks GenAI package"""

from setuptools import find_packages, setup

# pylint: disable-next=exec-used,consider-using-with
exec(open('databricks/model_training/version.py', 'r', encoding='utf-8').read())

install_requires = [
    'backoff>=2.2.1',
    'databricks-sdk>=0.29.0,<0.40',
    'datasets>=2.16.1,<=3.2',
    'gql[websockets]>=3.4.0',
    'mlflow>=2.14.2,<=2.19',
    'mosaicml-cli>=0.6',
    'requests>=2.26.0,<3',
    'typing_extensions>=4.7.1',
    'IPython>=8,<9',
    'pydantic>=2',
    'databricks-agents==0.13.0',
    'tenacity>=9.0.0',
    'rich>=12.6.0',
    'freezegun>=1.5,<2',
    'openai>=1.56,<2',
    'rouge-score>=0.1.2',
    'evaluate>=0.4.3',
]

extra_deps = {}

extra_deps['spark'] = [
    'databricks-connect==16.0.0',
]

extra_deps['dev'] = [
    'isort>=5.9.3',
    'packaging>=21,<25',
    'parameterized==0.9',
    'pre-commit>=2.17.0',
    'pylint>=3.0.0',
    'pyright==1.1.256',
    'pytest>=7.4.0',
    'yapf>=0.40.0',
]

extra_deps['release'] = [
    'build>=0.10.0',
    'twine>=4.0.2',
]

extra_deps['notebook'] = ['ipywidgets>=8,<9']

extra_deps['all'] = set(dep for deps in extra_deps.values() for dep in deps)

setup(
    name='databricks-genai',
    version=__version__,  # type: ignore pylint: disable=undefined-variable
    author='Databricks',
    author_email='genai-eng-team@databricks.com',
    description='Interact with the Databricks Generative AI APIs in python',
    url='https://docs.mosaicml.com/projects/mcli/en/latest/finetuning/finetuning.html',  # TODO: update with DBX docs
    include_package_data=True,
    package_data={},
    packages=find_packages(exclude=['tests']),
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    install_requires=install_requires,
    extras_require=extra_deps,
    python_requires='>=3.10',
    ext_package='databricks_genai',
)



"""
Errors for the DatabricksGenAI package.
"""
import logging
import re
from http import HTTPStatus
from typing import Any, Dict, List, Optional

import requests

DEFAULT_MESSAGE = 'Unknown Error'
ERROR_AUTH_KEY_MISSING = 'No API key or auth token was found. ' \
    'Please make sure you have your Databricks credentials set-up correctly.'

logger = logging.getLogger(__name__)


class DatabricksGenAIError(Exception):
    """Base class for exceptions in this module."""

    def __init__(self, message: str = ''):
        self.message = message
        super().__init__(self.message)


class DatabricksModelTrainingRequestError(DatabricksGenAIError):
    """Exception raised for errors in the request.

    Attributes:
        message -- explanation of the error
    """


class DatabricksGenAIResponseError(DatabricksGenAIError):
    """Exception raised for errors in the response.

    Attributes:
        message -- explanation of the error
    """


class DatabricksGenAIConfigError(DatabricksGenAIError):
    """Exception raised for errors in the config.

    Attributes:
        message -- explanation of the error
    """


class MAPIException(DatabricksGenAIError):
    """Exceptions raised when a request to MAPI fails

    Args:
        status: The status code for the exception
        message: A brief description of the error
        description: An optional longer description of the error

    Details:
    MAPI responds to failures with the following status codes:
    - 400: The request was misconfigured or missing an argument. Double-check the API and try again
    - 401: User credentials were either missing or invalid. Be sure to set your API key before making a request
    - 403: User credentials were valid, but the requested action is not allowed
    - 404: Could not find the requested resource(s)
    - 409: Attempted to create an object with a name that already exists. Change the name and try again.
    - 500: Internal error in MAPI. Please report the issue
    - 503: MAPI or a subcomponent is currently offline. Please report the issue
    """
    status: HTTPStatus
    message: str
    description: Optional[str] = None

    def __init__(self, status: HTTPStatus, message: str = DEFAULT_MESSAGE, description: Optional[str] = None):
        super().__init__()
        self.status = status
        self.message = message
        self.description = description

    def __str__(self) -> str:
        error_message = f'Error {self.status.value}: {self.message}'

        if self.description:
            error_message = f'{error_message}. {self.description}'

        return error_message

    @classmethod
    def from_mapi_error_response(cls, error: Dict[str, Any]):
        """Initializes a new exception based on error dict from a MAPI response
        """
        extensions = error.get('extensions', {})
        code = extensions.get('code', HTTPStatus.INTERNAL_SERVER_ERROR)
        try:
            status = HTTPStatus(code)
        except ValueError:
            logger.debug(f'Unknown status code {code}. Setting to 500')
            status = HTTPStatus.INTERNAL_SERVER_ERROR

        message = error.get('message', DEFAULT_MESSAGE)

        # TODO: could potentially include extensions['stacktrace'] as description for 500s internally
        # From apollo docs, this could only be available in dev?

        # Optionally translate to a more specific error, if one matches
        if RunConfigException.match(message):
            return RunConfigException(status=status, message=message)

        return MAPIException(status=status, message=message)

    @classmethod
    def from_bad_response(cls, response: requests.Response):
        return MAPIException(
            status=HTTPStatus(response.status_code),
            message=response.reason,
        )

    @classmethod
    def from_requests_error(cls, error: requests.exceptions.RequestException):
        """Initializes a new exception based on a requests RequestException
        """
        msg = 'Unable to connect to MAPI'
        if error.args:
            con = error.args[0]
            try:
                # Try to get the destination we tried to connect to
                # if the app is fully not accessible
                source = f'http://{con.pool.host}:{con.pool.port}{con.url}'
            except AttributeError:
                pass
            else:
                msg = f'{msg} at {source}'
        return MAPIException(status=HTTPStatus.SERVICE_UNAVAILABLE, message=msg)


class MultiMAPIException(MAPIException):
    """Raises 1 or more MAPI Exceptions

    Graphql can technically return multiple errors in the response. This
    allows the user to see all of them at once rather than having to debug
    one by one
    """

    def __init__(self, errors: List[MAPIException]) -> None:
        self.errors = errors
        status = max(e.status for e in errors)
        super().__init__(status)

    def __str__(self) -> str:
        return '\n'.join([str(x) for x in self.errors])


class RunConfigException(MAPIException):
    """Thrown when a run could not be created due to an incomplete FinalRunConfig
    """
    MATCH_MESSAGE = 'Bad run request'
    FIELD_PATTERN = re.compile('([A-Za-z]+) is a required field')

    def __init__(self, status: HTTPStatus, message: str = DEFAULT_MESSAGE, description: Optional[str] = None):
        super().__init__(status, message, description)
        fields = re.findall(self.FIELD_PATTERN, self.message)

        # Translate fields to make sense to the user
        fields_string = ', '.join(RunConfigException.translate_fields(fields))
        if fields:
            self.message = f'Run configuration is missing the following required values: {fields_string}'
        else:
            self.message = message

    @staticmethod
    def translate_fields(fields: List[str]) -> List[str]:
        # pylint: disable-next=import-outside-toplevel,cyclic-import
        from mcli.models.run_config import FinalRunConfig

        # pylint: disable-next=protected-access
        return [FinalRunConfig._property_translations.get(f, f) for f in fields]

    @classmethod
    def match(cls, message: str) -> bool:
        """Returns True if the error message suggests a RunConfigException
        """
        return RunConfigException.MATCH_MESSAGE in message


class ValidationError(DatabricksGenAIError):
    """Base class for interactive validation errors
    """


MAPI_DESERIALIZATION_ERROR = MAPIException(
    status=HTTPStatus.INTERNAL_SERVER_ERROR,
    message='Unknown issue deserializing data',
)



"""Holds the state of a KIE experiment"""
from typing import Dict, Optional, Type, Union

from mlflow.entities import Experiment
from pydantic import BaseModel, ConfigDict
from pyspark.sql.connect.dataframe import DataFrame as ConnectDataFrame
from pyspark.sql.dataframe import DataFrame as SparkDataFrame

from databricks.kie.prompt_builder import PromptBuilder

DataFrame = Union[SparkDataFrame, ConnectDataFrame]


class KIEState(BaseModel):
    """Holds the state of a KIE experiment"""
    experiment: Experiment  # The experiment to use for training
    ground_truth_prompt: str  # The prompt to use for ground truth generation

    prompt_builder: PromptBuilder
    zeroshot_prompt: str  # The prompt to use for zero-shot learning
    fewshot_prompt: Optional[str]  # The prompt to use for few-shot learning

    grounding_table_path: str  # Path to the grounding table
    val_table_path: str  # Path to the validation table
    train_jsonl_path: str  # Dataset path for training
    schema_path: str  # Path to the schema

    requires_grounding: bool  # Whether grounding is required
    requires_val: bool  # Whether validation is required
    requires_train_gen: bool  # Whether training data generation is required

    unlabeled_split_df: DataFrame  # Unlabeled data
    labeled_split_df: Optional[DataFrame]  # Labeled data
    grounding_df: Optional[DataFrame]  # Grounding data
    val_df: Optional[DataFrame]  # Validation data
    model_dfs: Optional[Dict[str, DataFrame]]  # Run-specific dataframes with eval data

    response_format: Type[BaseModel]  # The response format for the task

    num_grounding_samples: int  # Number of grounding samples
    num_val_samples: int  # Number of validation samples
    num_fewshot_samples: int  # Number of few-shot samples

    model_config = ConfigDict(
        protected_namespaces=(),
        arbitrary_types_allowed=True,
    )



"""Base runner for the KIE notebook"""
import base64
import copy
import json
import sys
import zlib
from typing import Dict, List, NamedTuple, Optional, Type, cast

import mlflow
from mlflow.entities import Run as MlflowRun
from pydantic import BaseModel
from pyspark.sql.dataframe import DataFrame

from databricks.kie.inference_utils import format_ai_query
from databricks.kie.kie_schema import ModelFactory
from databricks.kie.kie_state import KIEState
from databricks.kie.prompt_builder import PromptBuilder
from databricks.kie.steps.grounding import ground
from databricks.kie.steps.initialize import _SHARE_KEY, TaskInitializer, get_mlflow_experiment_link
from databricks.kie.steps.optimize import (SweepEvaluator, TrainingDataGenerator, ValidationDataGenerator,
                                           eval_baseline_models, launch_sweep)
from databricks.kie.steps.query_output import (get_prompts_from_run, prepare_output_table, ready_pt_endpoint_from_model,
                                               ready_pt_endpoint_from_run)
from databricks.kie.task_spec import KIETaskSpec
from databricks.model_serving.types.pt_endpoint import ProvisionedThroughputEndpoint
from databricks.model_serving.utils import get_model_specs
from databricks.model_training.api.utils import get_display

_DEFAULT_VIEW_DICT = {
    'searchFilter': '',
    'orderByKey': 'metrics.`metric/kie/overall_score/average`',
    'orderByAsc': False,
    'startTime': 'ALL',
    'lifecycleFilter': 'Active',
    'datasetsFilter': [],
    'modelVersionFilter': 'All Runs',
    'selectedColumns': [
        'attributes.`Source`', 'attributes.`Models`', 'attributes.`Dataset`', 'metrics.`overall_score`', 'params.`cost`'
    ],
    'runsExpanded': {},
    'runsPinned': [],
    'runsHidden': [],
    'runsVisibilityMap': {},
    'runsHiddenMode': 'FIRST_10_RUNS',
    'viewMaximized': False,
    'runListHidden': False,
    'isAccordionReordered': False,
    'useGroupedValuesInCharts': True,
    'hideEmptyCharts': True,
    'groupBy': None,
    'groupsExpanded': {},
    'autoRefreshEnabled': True,
    'globalLineChartConfig': {
        'xAxisKey': 'step',
        'lineSmoothness': 0,
        'selectedXAxisMetricKey': ''
    },
    'metricAggregateTypeMap': {},
}


class KIEOutput(NamedTuple):
    endpoint: ProvisionedThroughputEndpoint
    ai_query: str


class KIERunner:
    """Runner class to be used in the KIE notebook"""
    task_spec: KIETaskSpec
    state: KIEState

    num_grounding_samples: int = 10
    num_val_samples: int = 25
    num_fewshot_samples: int = 3

    grounding_model_name = "balanced"
    optimized_model_names = ["cost-optimized-medium", "cost-optimized-small"]

    def __init__(self, task_spec: KIETaskSpec):
        self.task_spec = task_spec
        self.mlflow_client = mlflow.MlflowClient()
        initializer = TaskInitializer(task_spec)
        self.state = initializer.initialize(num_grounding_samples=self.num_grounding_samples,
                                            num_val_samples=self.num_val_samples,
                                            num_fewshot_samples=self.num_fewshot_samples)

        metrics_to_view = [f'metric/kie/{k}/average' for k in self.state.response_format.model_fields.keys()]
        self._set_share_key_tag(self.state.experiment.experiment_id, _SHARE_KEY, metrics_to_view)

        model_specs = get_model_specs()
        self.grounding_model = model_specs[self.grounding_model_name]
        assert self.grounding_model.endpoint is not None, "Grounding model must have an endpoint defined"
        assert self.grounding_model.uc_schema is not None, "Grounding model must have a UC schema path defined"
        self.optimized_models = [model_specs[name] for name in self.optimized_model_names]

    @staticmethod
    def _get_share_key_value(metrics_to_display: list[str],) -> str:
        view_dict = copy.deepcopy(_DEFAULT_VIEW_DICT)

        view_dict['selectedColumns'] += [f'metrics.`{m}`' for m in metrics_to_display] + [view_dict['orderByKey']]

        new_value = 'deflate;' + base64.b64encode(zlib.compress(json.dumps(view_dict).encode())).decode()

        return new_value

    def _set_share_key_tag(
        self,
        experiment_id: str,
        share_key: str,
        metrics_to_display: list[str],
    ) -> None:
        new_value = self._get_share_key_value(metrics_to_display)
        self.mlflow_client.set_experiment_tag(experiment_id=experiment_id,
                                              key=f'mlflow.sharedViewState.{share_key}',
                                              value=new_value)

    @property
    def response_format(self) -> Type[BaseModel]:
        return self.state.response_format

    @response_format.setter
    def response_format(self, response_format: Type[BaseModel]):
        self.recompile(response_format)

    @property
    def base_model_cost(self) -> float:
        return self.grounding_model.cost_per_hour

    @property
    def base_model_uc_schema(self) -> str:
        return self.grounding_model.uc_schema

    @property
    def ai_query(self) -> str:
        if not self.state.ai_query:
            raise ValueError("Please run `ready_ai_query_output` first")
        return self.state.ai_query

    def recompile(self, response_format: Type[BaseModel]):
        self.state.response_format = response_format

        # Cache the response format
        ModelFactory.to_file(self.state.response_format, self.state.schema_path)

        # Rebuild prompts
        self.state.prompt_builder = PromptBuilder(self.state.response_format)
        self.state.ground_truth_prompt = self.state.prompt_builder.build_prompt()
        self.state.zeroshot_prompt = self.state.prompt_builder.build_prompt(include_markdown=True)

        # Request that grounding be re-run
        self.state.requires_grounding = True

        metrics_to_view = [f'metric/kie/{k}/average' for k in self.state.response_format.model_fields.keys()]
        self._set_share_key_tag(self.state.experiment.experiment_id, _SHARE_KEY, metrics_to_view)

    def ground(self) -> DataFrame:
        ground(self.task_spec, self.state, self.grounding_model)

        # Update grounding cache
        assert self.state.grounding_df is not None  # We just grounded
        self.state.grounding_df.write.mode("overwrite").saveAsTable(self.state.grounding_table_path)

        return cast(DataFrame, self.state.grounding_df)

    def optimize(self, _gen_with_pt: bool = True) -> Dict[str, DataFrame]:
        # Generate validation data
        path = get_mlflow_experiment_link(self.state.experiment.experiment_id)
        print("Kicking off model optimizations. This step may take a while. " +
              f"You can track the progress at your MLflow experiment:\n{path}")

        val_generator = ValidationDataGenerator(self.task_spec,
                                                self.state,
                                                self.grounding_model,
                                                use_provisioned_throughput=_gen_with_pt)
        val_generator.generate()

        assert self.state.val_df is not None
        self.state.val_df.write.mode("overwrite").saveAsTable(self.state.val_table_path)

        # Feedback to user
        print("We'll test a few models using the validation dataset:\n")
        display = get_display()
        display(self.state.val_df.select("request", "expected_response").limit(10))  # TODO: This needs to flush somehow
        sys.stdout.flush()  # Maybe this will help?

        # Eval baseline models
        eval_baseline_models(self.task_spec, self.state, self.optimized_models)

        # Generate training data
        train_generator = TrainingDataGenerator(self.task_spec,
                                                self.state,
                                                self.grounding_model,
                                                self.optimized_models,
                                                use_provisioned_throughput=_gen_with_pt)
        train_generator.generate()

        # Train generator may have not actually created training data due to not enough docs
        if not self.state.requires_train_gen:
            # Launch the sweep
            launch_sweep(self.task_spec, self.state, self.optimized_models)

            # Follow sweep runs and eval models
            sweep_evaluator = SweepEvaluator(self.task_spec, self.state)
            sweep_evaluator.evaluate_sweep()

        assert self.state.model_dfs is not None
        return cast(Dict[str, DataFrame], self.state.model_dfs)

    def follow_optimizations(self) -> Dict[str, DataFrame]:
        sweep_evaluator = SweepEvaluator(self.task_spec, self.state)
        sweep_evaluator.evaluate_sweep()

        assert self.state.model_dfs is not None
        return cast(Dict[str, DataFrame], self.state.model_dfs)

    def ready_ai_query_output(self, use_baseline: bool = False, run_id: Optional[str] = None) -> KIEOutput:
        """Ready a PT endpoint and AI query command for the chosen model
        
        The PT endpoint will be created with "scale-to-zero" enabled, so it will automatically
        scale down when not in use.
        
        Args:
            use_baseline: Whether to use the baseline model
            run_id: The MLflow run ID of the model to use
        
        Returns:
            KIEOutput: A named tuple containing the `endpoint` and `ai_query` command
        
        Raises:
            ValueError: If neither `use_baseline` nor `run_id` is provided or if the run
                does not have all required artifacts
        """

        # Prepare the output table
        prepare_output_table(self.task_spec, self.state)

        # Stand-up an endpoint for the chosen model
        if use_baseline:
            endpoint = ready_pt_endpoint_from_model(self.base_model_uc_schema)
            system_prompt = trailing_prompt = self.state.zeroshot_prompt
        elif run_id:
            run = mlflow.get_run(run_id)
            endpoint = ready_pt_endpoint_from_run(run)
            system_prompt, trailing_prompt = get_prompts_from_run(run)
        else:
            raise ValueError("Please provide a run_id or set use_baseline=True")

        # Format the AI query command
        ai_query = format_ai_query(self.task_spec.output_table,
                                   endpoint,
                                   system_prompt=system_prompt,
                                   trailing_prompt=trailing_prompt)

        return KIEOutput(endpoint, ai_query)

    def choose_best_run(self) -> MlflowRun:
        # Get the best run from the experiment
        runs = mlflow.search_runs(experiment_ids=[self.state.experiment.experiment_id],
                                  filter_string="tags.registered_to LIKE '%'",
                                  order_by=["metrics.overall_score DESC"],
                                  max_results=1,
                                  output_format="list")
        runs = cast(List[MlflowRun], runs)
        if not runs:
            raise RuntimeError("Found no runs in the experiment that have been scored")
        return runs[0]



""" Finetune config """
import logging
from dataclasses import asdict, dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

import yaml
from mcli.utils.utils_config import (BaseSubmissionConfig, ExperimentTrackerConfig, ExperimentTrackerTranslation,
                                     strip_nones)
from mcli.utils.utils_string_functions import camel_case_to_snake_case, snake_case_to_camel_case

from databricks.model_training.api.exceptions import DatabricksGenAIConfigError

logger = logging.getLogger(__name__)


class TrainTaskType(Enum):
    """ The Type of Finetune Task to run """
    INSTRUCTION_FINETUNE = 'INSTRUCTION_FINETUNE'
    CONTINUED_PRETRAIN = 'CONTINUED_PRETRAIN'
    CHAT_COMPLETION = 'CHAT_COMPLETION'


@dataclass
class TrainConfig(BaseSubmissionConfig):
    """Input for training run

    Args:
        model (str): Model to finetune (e.g. 'mosaicml/mpt-30b')
        task_type (str): Task type to finetune on
        train_data_path (str): HF dataset or remote path to training data
        save_folder (str): Folder to save checkpoints and HF checkpoints, currently must be an s3 or GCP path
        eval_data_path (Optional[str]): HF dataset or remote path to eval data
        eval_prompts (Optional[List[str]]): A list of prompt strings to generate from periodically during training
        custom_weights_path (Optional[str]): Remote location of custom weights to use for model training
        training_duration (str): Composer variable for the total number of epochs or tokens to train on
            (e.g. 2ep or 10000tok) [TO BE DEPRECATED]
        training_durations (Optional[List[str]]): List of epochs or tokens to train for
        experiment_tracker (Optional[ExperimentTrackerConfig]): Experiment tracker config
        learning_rate (Optional[float]): Learning rate to use [TO BE DEPRECATED]
        learning_rates (Optional[List[float]]): List of learning rates to use
        context_length (Optional[int]): Context length of the model. Override only works for MPT model families
        disable_credentials_check (Optional[bool]): Flag to disable checking credentials (S3, Databricks, etc.)
            on finetune submission
    """

    model: str
    train_data_path: str
    save_folder: str = 'dbfs:/databricks/mlflow-tracking/{mlflow_experiment_id}/{mlflow_run_id}/artifacts'
    task_type: Optional[str] = 'INSTRUCTION_FINETUNE'
    eval_data_path: Optional[str] = None
    eval_prompts: Optional[List[str]] = None
    custom_weights_path: Optional[str] = None
    training_duration: Optional[str] = None
    training_durations: Optional[List[str]] = None
    experiment_tracker: Optional[ExperimentTrackerConfig] = None
    learning_rate: Optional[float] = None
    learning_rates: Optional[List[float]] = None
    context_length: Optional[int] = None
    # The data prep config is an internal config passed to MAPI to define requirements for data transformation.
    # The SDK only minimally exposes the `data_prep_cluster_id` field to the user, which we will pack and unpack from
    # the `data_prep_config` automatically.
    data_prep_config: Optional[Dict] = None
    disable_credentials_check: Optional[bool] = None

    _required_properties = {'model', 'train_data_path', 'save_folder'}
    _finetune_properties = {
        'model', 'train_data_path', 'save_folder', 'eval_data_path', 'eval_prompts', 'custom_weights_path',
        'training_duration', 'training_durations', 'experiment_tracker', 'learning_rate', 'learning_rates',
        'context_length', 'task_type', 'data_prep_config', 'disable_credentials_check'
    }

    def __str__(self) -> str:
        return yaml.safe_dump(asdict(self))

    @classmethod
    def validate_dict(cls, dict_to_use: Dict, show_unused_warning: bool = True) -> Dict:
        """Load the config from the provided dictionary.
        """
        config_dict = strip_nones(dict_to_use)
        unused_keys = set(config_dict) - cls._finetune_properties
        for key in unused_keys:
            del config_dict[key]
        missing = cls._required_properties - set(dict_to_use)
        if missing:
            raise DatabricksGenAIConfigError(f'Missing required fields: {", ".join(missing)}',)
        if len(unused_keys) > 0 and show_unused_warning:
            raise DatabricksGenAIConfigError(
                f'Encountered unknown fields in training config: {", ".join(unused_keys)}.')
        return config_dict

    @classmethod
    def from_dict(cls, dict_to_use: Dict[str, Any], show_unused_warning: bool = True):
        """Load the config from the provided dictionary.
        """
        config_dict = cls.validate_dict(dict_to_use, show_unused_warning)
        return cls(**config_dict)

    @classmethod
    def from_mapi_response(cls, response: Dict[str, Any]):
        """Load the config from mapi response.
        """
        experiment_tracker_result = {}
        for key, value in response.items():
            if key == 'experimentTracker':
                experiment_tracker_result['experiment_tracker'] = ExperimentTrackerTranslation.from_mapi(value)
            else:
                experiment_tracker_result[camel_case_to_snake_case(key)] = value
        return cls.from_dict(experiment_tracker_result)

    def to_create_api_input(self) -> Dict[str, Dict[str, Any]]:
        """Converts the TrainConfig object to Compute and Finetune that can be used to create a finetune run
        """
        finetune = {}
        config_dict = strip_nones(self.__dict__)
        for key, val in config_dict.items():
            if key == 'experiment_tracker':
                finetune['experimentTracker'] = ExperimentTrackerTranslation.to_mapi(val)
                continue
            translated_key = snake_case_to_camel_case(key)
            if key in self._finetune_properties:
                finetune[translated_key] = val

        return {'config': finetune}

    def to_yaml(self) -> str:
        """Converts the TrainConfig object to a yaml string with only used keys
        """
        filtered_config = {k: v for k, v in self.__dict__.items() if v}
        return yaml.safe_dump(filtered_config)

    def __post_init__(self):
        if self.task_type not in TrainTaskType.__members__:
            raise DatabricksGenAIConfigError(f'Invalid task_type: {self.task_type}, currently '
                                             f'supported options are: {", ".join([x.value for x in TrainTaskType])}')



"""Cancel a model training run"""

from typing import Any, Dict, List, Optional, Union

from databricks.model_training.api.engine import get_return_response, run_plural_mapi_request
from databricks.model_training.api.exceptions import DatabricksModelTrainingRequestError
from databricks.model_training.types.training_run import TrainingRun

QUERY_FUNCTION = 'stopFinetunes'
VARIABLE_DATA_NAME = 'getFinetunesData'
OPTIONAL_DATA_NAME = 'stopFinetunesData'
QUERY = f"""
mutation StopFinetunes(${VARIABLE_DATA_NAME}: GetFinetunesInput!, ${OPTIONAL_DATA_NAME}: StopFinetunesInput) {{
  {QUERY_FUNCTION}({VARIABLE_DATA_NAME}: ${VARIABLE_DATA_NAME}, {OPTIONAL_DATA_NAME}: ${OPTIONAL_DATA_NAME}) {{
    id
    name
    status
    createdById
    createdByEmail
    createdAt
    updatedAt
    startedAt
    completedAt
    reason
    isDeleted
  }}
}}"""


def cancel(
    runs: Optional[Union[str, TrainingRun, List[str], List[TrainingRun]]] = None,
    experiment_id: Optional[str] = None,
) -> int:
    """
    Cancel a training run, list of training runs, or all runs under a single MLflow experiment id without
    deleting them. If the run does not exist or if the run has already terminated, an error will be raised.

    Args:
        runs (Optional[Union[str, TrainingRun, List[str], List[TrainingRun]]]): The
            training run(s) to cancel. Can be a single run or a list of runs.
        experiment_id (Optional[str]): The MLflow experiment ID to cancel by.
            This will cancel all training runs under the given experiment.

    Raises:
        DatabricksModelTrainingRequestError: Raised if stopping any of the requested runs failed

    Returns:
        int: The number of training runs cancelled
    """

    if not (runs or experiment_id):
        raise DatabricksModelTrainingRequestError('Must provide training run(s) or MLflow experiment ID to cancel')

    if runs and experiment_id:
        raise DatabricksModelTrainingRequestError(
            'Cannot provide both training run(s) and MLflow experiment ID to cancel. Please specify one or the other.')

    filters = {}
    runs_list = []
    if experiment_id:
        filters['mlflowExperimentId'] = {'equals': experiment_id}
        cancel_msg = 'MLflow experiment ID'
    else:
        if isinstance(runs, (str, TrainingRun)):
            runs_list = [runs]
        elif runs is not None:
            runs_list = runs
        training_run_names = [run.name if isinstance(run, TrainingRun) else run for run in runs_list]
        if training_run_names:
            filters['name'] = {'in': training_run_names}
        cancel_msg = 'training run(s)'

    variables: Dict[str, Dict[str, Any]] = {
        VARIABLE_DATA_NAME: {
            'filters': filters
        },
        OPTIONAL_DATA_NAME: {
            'reason': 'Canceled by user.'
        }
    }

    try:
        response = run_plural_mapi_request(
            query=QUERY,
            query_function=QUERY_FUNCTION,
            return_model_type=TrainingRun,
            variables=variables,
        )
        canceled_runs = get_return_response(response)
        num_deleted = len(canceled_runs)
        if num_deleted == 0:
            print('No active runs found to cancel.')
        else:
            canceled_run_names = ', '.join([run.name for run in canceled_runs])
            not_found = ''
            if num_deleted < len(runs_list):
                not_found = f' {len(runs_list) - num_deleted} run(s) were not found or already terminated.'
            print(f'Canceled {canceled_run_names} successfully.{not_found}')
        return num_deleted
    except Exception as e:
        raise DatabricksModelTrainingRequestError(
            f'Failed to cancel {cancel_msg} {runs if runs else experiment_id}. Please make sure the run \
                                                  has not completed or failed and try again.') from e



"""Databricks Gen AI Package, aliased as databricks"""
# https://packaging.python.org/guides/packaging-namespace-packages/#pkgutil-style-namespace-packages
# This file should only contain the following line. Otherwise other sub-packages databricks.* namespace
# may not be importable.
__path__ = __import__('pkgutil').extend_path(__path__, __name__)

try:
    import pyspark
    del pyspark
except ImportError as e:
    raise ImportError("PySpark is not installed. Please install databricks-connect to use this package.") from e



"""Utility functions for the API."""

import hashlib
import logging
import os
import re
from typing import Any, Callable, List, NamedTuple, Optional, Tuple, cast

from databricks.sdk import WorkspaceClient
from IPython import get_ipython  # type: ignore
from pyspark.sql.session import SparkSession

logger = logging.getLogger(__name__)

_LOCAL_DEV_CONFIG_PROFILE = 'DBX_DEV_LOCAL'
_DATABRICKS_CONFIG_PROFILE_ENV = 'DATABRICKS_CONFIG_PROFILE'
_TEST_CONFIG_PROFILE = 'DBX_DEV_TEST'


def get_browser_url() -> str:
    w = WorkspaceClient()
    hostname = w.dbutils.notebook.entry_point.getDbutils().notebook().getContext().browserHostName().get()
    return f"https://{hostname}"


def get_spark() -> SparkSession:
    if is_running_in_databricks_notebook():
        try:
            from databricks.connect import DatabricksSession  # pylint: disable=import-outside-toplevel
        except ImportError as e:
            raise ImportError(
                "Databricks Connect is not installed. Please install databricks-connect to use this package.") from e

        # This is a hack - typing should be fine and hopefully will work better when in a notebook
        return cast(SparkSession, DatabricksSession.builder.getOrCreate())
    return SparkSession.builder.getOrCreate()


def is_running_in_databricks_notebook() -> bool:
    try:
        shell = get_ipython().__class__.__name__  # pylint: disable = undefined-variable
        if shell == 'DatabricksShell':
            return True
        return False
    except Exception:  # pylint: disable = broad-except
        return False


class NotebookDetails(NamedTuple):
    """Details about the current notebook."""
    notebook_path: str
    notebook_id: str


def get_current_notebook_details() -> Optional[NotebookDetails]:
    if not is_running_in_databricks_notebook():
        return
    w = WorkspaceClient()
    context = w.dbutils.notebook.entry_point.getDbutils().notebook().getContext()
    return NotebookDetails(context.notebookPath().get(), context.notebookId().get())


def normalize_table_name(table_name: str) -> str:
    """Normalize SQL table name by adding backticks where needed.
    
    Args:
        table_name: Raw table name or schema.table name
        
    Returns:
        Normalized table name with backticks where needed
    """

    if table_name.startswith('`') and table_name.endswith('`'):
        return table_name
    table_name_parts = table_name.split('.')
    return '.'.join([f"`{part.strip('`')}`" if '-' in part else part for part in table_name_parts])


def table_schema_overlaps(table_name: str, protected_columns: List[str]) -> List[str]:
    """Check if the table already has columns that overlap with the schema that is "protected";
       If the customer has these columns, our code will run into issues.
    Args:
        table_name: The name of the table to check.
        protected_columns: The columns to check for overlap with.
    Returns:
        List[str]: The columns that overlap with the schema that we want to use for the AI query.
    """
    spark = get_spark()
    df = spark.read.table(table_name)
    overlapping_columns = []
    for column in protected_columns:
        if column in df.columns:
            overlapping_columns.append(column)
    return overlapping_columns


def format_table_name(name: str) -> str:
    """Formats a valid table name based on the provided string. 
    
    Replaces all invalid characters with _.
    
    Args:
        name: The name to convert to a valid table name.
        
    Returns:
        A valid table name.
    """
    # Convert to lowercase first
    name = name.lower()

    # Replace any character that isn't alphanumeric or underscore with underscore
    name = re.sub(r'[^a-z0-9_]', '_', name)

    return name


def get_schema_from_table(table_name: str) -> str:
    schema = table_name.replace('`', '').split('.')[:-1]
    return normalize_table_name('.'.join(schema))


def check_if_table_exists(table_name: str) -> bool:
    """Checks if the specified table already exists
    
    Args:
        table_name: The name of the table to validate.
    
    Returns:
        True if the table exists, False otherwise.
    """
    w = WorkspaceClient()
    r = w.tables.exists(normalize_table_name(table_name))
    return bool(r.table_exists)


def check_table_has_columns(table_name: str, columns: Tuple[str]) -> bool:

    spark = get_spark()
    df = spark.read.table(normalize_table_name(table_name))
    for column in columns:
        if column not in df.columns:
            return False
    return True


def get_display() -> Callable[[Any], None]:
    from IPython.display import display as default_display  # pylint: disable = import-outside-toplevel
    if is_running_in_databricks_notebook():
        # Get display from user namespace
        display = get_ipython().user_ns.get('display', default_display)
        return display
    else:
        return default_display


def get_me() -> str:
    """
    Get who is currently logged in.

    Returns:
        str: The name of the current user.
    """
    # TODO remove, only used for testing
    if os.environ.get(_DATABRICKS_CONFIG_PROFILE_ENV, '') in [
            _LOCAL_DEV_CONFIG_PROFILE,
            _TEST_CONFIG_PROFILE,
    ]:
        return 'me'

    w = WorkspaceClient()
    me = w.current_user.me().user_name or ''
    if not me:
        raise EnvironmentError('Could not determine the current user. Please check your environment.')
    logger.debug(f'You are {me}')
    return me


def get_workspace() -> int:
    """
    Get the current workspace id
    
    Returns:
        int: The workspace ID
    """
    w = WorkspaceClient()
    if is_running_in_databricks_notebook():
        ctx = w.dbutils.entry_point.getDbutils().notebook().getContext()
        return int(ctx.workspaceId().get())
    else:
        # not seeing this property on the object but it really does exist as far as I can tell
        return w.get_workspace_id()  # type: ignore


def get_host_and_token_from_env() -> Tuple[str, str]:
    """
    Get the host and token from the environment

    In a databricks notebook, the host will be the underlying region and env (i.e. oregon.staging.databricks.com)
    """
    if os.environ.get(_DATABRICKS_CONFIG_PROFILE_ENV, '') == _TEST_CONFIG_PROFILE:
        return 'test', 'test'
    w = WorkspaceClient()
    if is_running_in_databricks_notebook():
        ctx = w.dbutils.entry_point.getDbutils().notebook().getContext()
        host = ctx.apiUrl().get()
        token = ctx.apiToken().get()
    else:
        host = w.config.host
        token = w.config.token
    if not host or not token:
        raise EnvironmentError('Could not find root URL and/or token. Please check your environment.')
    return host, token


def get_cluster_id() -> str:
    """
    Get the cluster ID from the environment
    """
    if os.environ.get(_DATABRICKS_CONFIG_PROFILE_ENV, '') == _TEST_CONFIG_PROFILE:
        return 'test'
    w = WorkspaceClient()
    if is_running_in_databricks_notebook():
        ctx = w.dbutils.notebook.entry_point.getDbutils().notebook().getContext()
        cluster_id = ctx.clusterId().get()
    else:
        cluster_id = w.config.cluster_id
    return cluster_id


def md5_hash(s: str, num_digits=8) -> str:
    return hashlib.md5(s.encode()).hexdigest()[:num_digits]



"""Utils for working with model serving"""

from typing import NamedTuple, Optional

from mlflow import MlflowClient

MODEL_CONTEXT_LENGTH = 131072


class ModelSpecs(NamedTuple):
    name: str
    endpoint: Optional[str]
    cost_per_hour: float
    ft_model_name: str
    is_hosted: bool = True
    uc_schema: Optional[str] = None
    version: Optional[str] = None
    cost_per_m_ft_tokens: Optional[float] = None


def get_latest_model_version(model_name: str) -> str:
    """Get the latest version of a model using Mlflow

    Args:
        model_name (str): The name of the model

    Returns:
        str: The latest version of the model
    """
    client = MlflowClient()
    model_versions = client.search_model_versions(f"name='{model_name}'")

    if not model_versions:
        raise ValueError(f"No model versions found for model '{model_name}'")

    version_numbers = [int(model.version) for model in model_versions]
    most_recent_version = max(version_numbers)
    return str(most_recent_version)


def get_model_by_ft_name(ft_model_name: str) -> Optional[str]:
    model_specs = get_model_specs()
    for model_spec in model_specs.values():
        if model_spec.ft_model_name == ft_model_name:
            return model_spec.name


def get_model_specs():
    model_specs = {
        'balanced':
            ModelSpecs("balanced",
                       endpoint='databricks-meta-llama-3-3-70b-instruct',
                       cost_per_hour=24,
                       ft_model_name='meta-llama/Llama-3.3-70B-Instruct',
                       uc_schema='system.ai.llama_v3_3_70b_instruct',
                       version=get_latest_model_version('system.ai.llama_v3_3_70b_instruct')),
        'cost-optimized-medium':
            ModelSpecs("cost-optimized-medium",
                       endpoint=None,
                       cost_per_hour=7.42,
                       ft_model_name='meta-llama/Meta-Llama-3.1-8B-Instruct',
                       is_hosted=False,
                       uc_schema='system.ai.meta_llama_v3_1_8b_instruct',
                       version=get_latest_model_version('system.ai.meta_llama_v3_1_8b_instruct'),
                       cost_per_m_ft_tokens=4),
        'cost-optimized-small':
            ModelSpecs("cost-optimized-small",
                       endpoint=None,
                       cost_per_hour=6.50,
                       ft_model_name='meta-llama/Llama-3.2-3B-Instruct',
                       is_hosted=False,
                       uc_schema='system.ai.llama_v3_2_3b_instruct',
                       version=get_latest_model_version('system.ai.llama_v3_2_3b_instruct'),
                       cost_per_m_ft_tokens=2.50),
    }
    return model_specs



"""Defines the task spec for KIE"""
import os
import re
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, model_validator

from databricks.kie.data_utils import IGNORED_CACHE_ROOT, get_valid_files
from databricks.model_training.api.exceptions import ValidationError
from databricks.model_training.api.utils import (check_if_table_exists, check_table_has_columns, format_table_name,
                                                 get_me, get_schema_from_table, get_spark, normalize_table_name,
                                                 table_schema_overlaps)
from databricks.model_training.api.validation import (validate_delta_table, validate_uc_path_write_permissions,
                                                      validate_uc_permissions)


def get_default_experiment_name_with_suffix(suffix: str) -> str:
    """Returns a default experiment name with a suffix.

    Args:
        suffix: Suffix to append to the default experiment name

    Returns: Experiment name in format /Users/{username}/{suffix} .
    """
    username = get_me()
    return f"/Users/{username}/{suffix}"


class KIETaskSpec(BaseModel):
    """Task spec for KIE experiments"""
    unlabeled_dataset: str
    json_examples: List[Dict[str, Any]]
    experiment_name: str
    labeled_dataset: Optional[str]
    labeled_dataset_text_column: str
    labeled_dataset_output_json_column: str
    output_path: str
    output_table: str

    @staticmethod
    def _get_default_output_path(unlabeled_dataset: str, experiment_name: str) -> str:
        # Default to a cache folder in the dataset
        root_name = format_table_name(os.path.basename(experiment_name))
        return os.path.join(unlabeled_dataset, f"{IGNORED_CACHE_ROOT}{root_name}")

    @staticmethod
    def _get_default_output_table(output_path: str, unlabeled_dataset: str) -> str:
        # Default to a table in the same schema as the input files
        output_schema = ".".join(output_path.split("/")[2:4])
        dataset_name = format_table_name(os.path.basename(unlabeled_dataset))
        return f"{output_schema}.{dataset_name}"

    @staticmethod
    def _get_default_experiment_name(unlabeled_dataset: str) -> str:
        # Choose a default experiment name
        # Extract dataset name from full path and normalize it
        unlabeled_dataset = os.path.basename(unlabeled_dataset.rstrip("/").replace("_", "-").lower())
        # Include user path
        return get_default_experiment_name_with_suffix(f"kie-{unlabeled_dataset}")

    @classmethod
    def create(
        cls,
        unlabeled_dataset: str,
        json_examples: List[Dict[str, Any]],
        experiment_name: Optional[str],
        labeled_dataset: Optional[str],
        labeled_dataset_text_column: str,
        labeled_dataset_output_json_column: str,
        output_path: Optional[str],
        output_table: Optional[str],
    ) -> 'KIETaskSpec':
        # Normalize path
        unlabeled_dataset = unlabeled_dataset.rstrip("/")

        if experiment_name is None:
            # Choose a default experiment name based on the dataset name
            experiment_name = cls._get_default_experiment_name(unlabeled_dataset)

        if output_path is None:
            output_path = cls._get_default_output_path(unlabeled_dataset, experiment_name=experiment_name)

        if output_table is None:
            output_table = cls._get_default_output_table(output_path, unlabeled_dataset)

        return cls(
            unlabeled_dataset=unlabeled_dataset,
            json_examples=json_examples,
            experiment_name=experiment_name,
            labeled_dataset=labeled_dataset,
            labeled_dataset_text_column=labeled_dataset_text_column,
            labeled_dataset_output_json_column=labeled_dataset_output_json_column,
            output_path=output_path,
            output_table=normalize_table_name(output_table),
        )

    @model_validator(mode='after')
    def validate_spec(self: 'KIETaskSpec') -> 'KIETaskSpec':
        try:
            return validate_task_spec(self)
        except ValidationError as e:
            # Re-raise validation errors for pydantic
            raise ValueError(str(e)) from e


def validate_can_read_from_unlabeled_dataset(unlabeled_dataset: str):
    try:
        os.listdir(unlabeled_dataset)
    except OSError as e:
        raise ValueError(f"Could not read from unlabeled dataset at {unlabeled_dataset}. " +
                         "Please make sure the path is correct and you have read permissions.") from e


def validate_schema_starting_point(task_spec: KIETaskSpec):
    if not task_spec.json_examples and not task_spec.labeled_dataset:
        raise ValueError("Please provide example JSON outputs so we can know what to extract")


def validate_output_write_permissions(output_path: str):
    try:
        os.makedirs(output_path, exist_ok=True)
    except OSError as e:
        raise ValueError(f"Could not create output folder at {output_path}. " +
                         "Please make sure you have write permissions to the parent directory, " +
                         "or specify an `output_path` where you do") from e

    validate_uc_path_write_permissions(output_path)


def validate_valid_output_path(output_path: str, unlabeled_dataset: str):
    if output_path.rstrip("/") == unlabeled_dataset.rstrip("/"):
        raise ValueError(
            "Output path cannot be the same as the unlabeled dataset. Please provide a different output path")

    if "." in output_path:
        raise ValueError("Output path cannot contain a '.' character. Please provide a different output path")


def validate_output_table_name(output_table: str):
    names = output_table.split(".")
    if len(names) != 3:
        raise ValueError("Output table name must be in the format `catalog.schema.table`")
    for name in names:
        if re.match(r'^[a-zA-Z0-9_`-]+$', name) is None:
            raise ValueError("Output table name must only contain alphanumeric characters, underscores, and hyphens.")


def validate_labeled_dataset_is_valid(labeled_dataset: str, labeled_dataset_text_column: str,
                                      labeled_dataset_output_json_column: str):
    spark = get_spark()
    labeled_df = spark.read.table(labeled_dataset)
    schema = dict(labeled_df.dtypes)

    # Validate the provided input column
    if labeled_dataset_text_column not in schema:
        raise ValueError(f"The provided labeled_dataset_text_column {labeled_dataset_text_column} " +
                         "is not in the table. Please check your configuration.")

    # Validate the provided output column, if specified
    if labeled_dataset_output_json_column not in schema:
        raise ValueError(f"The provided labeled_dataset_output_json_column {labeled_dataset_output_json_column} " +
                         "is not in the table. Please check your configuration.")
    # If we have the following columns, we will have issues in other steps since we never isolate columns
    # response: (grounding.py:L29) The table will have two `response` columns which can't be saved
    # extracted_response: (inference_utils.py:L97) The table will have two `extracted_response` columns
    # raw_response: (inference_utils.py:L96) The table will have two `raw_response` columns
    # Note: Other columns like "rand", "row_id", etc that will be overwritten and can lead to edge case issues
    check_protected_columns(labeled_dataset, ["response", "extracted_response", "raw_response"])


def validate_row_count(
    unlabeled_dataset: str,
    required_count: int = 10,
    recommended_count: int = 1000,
) -> None:
    all_files = get_valid_files(unlabeled_dataset)
    row_count = len(all_files)
    if row_count < required_count:
        raise ValueError(f'Insufficient data. We require at least {required_count} unlabeled examples ' +
                         f'(recommend at least {recommended_count}). Found only {row_count} examples.')
    if row_count < recommended_count:
        print(f'Warning: Found {row_count} unlabeled examples, ' +
              f'we recommend at least {recommended_count} for best results.')


def check_protected_columns(table_name: str, protected_columns: List[str]):
    if overlapping_columns := table_schema_overlaps(table_name, protected_columns):
        raise ValueError(f"Table {table_name} already has the columns named {overlapping_columns}. "
                         "Please use different column names as this is a conflict.")


def validate_task_spec(task_spec: KIETaskSpec) -> KIETaskSpec:

    # Validate that we can read from the unlabeled_dataset
    validate_can_read_from_unlabeled_dataset(task_spec.unlabeled_dataset)

    # Validate that we have enough unlabeled data to do anything
    validate_row_count(task_spec.unlabeled_dataset)

    # Validate that we have some schema starting point
    validate_schema_starting_point(task_spec)

    # Validate that the output path is valid
    validate_valid_output_path(task_spec.output_path, task_spec.unlabeled_dataset)

    # Validate permission to write to output paths
    validate_output_write_permissions(task_spec.output_path)
    print(f"✔ Validated write permissions to output_path: {task_spec.output_path}")

    # Validate output table has valid characters
    validate_output_table_name(task_spec.output_table)

    # Validate we can write to the output schema
    output_schema = get_schema_from_table(task_spec.output_table)
    validate_uc_permissions(output_schema, 'schema', ['ALL_PRIVILEGES', 'USE_SCHEMA'], input_name='output_schema')

    # Ensure the output table doesn't already exist
    if check_if_table_exists(task_spec.output_table):
        # If output_table already exists, the user may be repeating KIE
        # Verify that it has a "request" column
        if not check_table_has_columns(task_spec.output_table, ("request",)):
            raise ValueError(f"Output table {task_spec.output_table} already exists, " +
                             "but does not seem to be from a KIE experiment. Please choose a different path.")

    print(f"✔ Validated write permissions to output_table: {task_spec.output_table}")

    # Validate the labeled dataset and ensure all columns specified exist
    if task_spec.labeled_dataset:
        # Validate that you can access the labeled_dataset
        validate_delta_table(task_spec.labeled_dataset, "labeled_dataset")

        # Validate the labeled dataset is valid
        validate_labeled_dataset_is_valid(task_spec.labeled_dataset, task_spec.labeled_dataset_text_column,
                                          task_spec.labeled_dataset_output_json_column)
        print(f"✔ Validated permissions and contents of labeled_dataset: {task_spec.labeled_dataset}")

    print("✔ Validation checks passed!")
    return task_spec
